
import pymongo
from pymongo import MongoClient

import uuid
import unittest
import pprint
from matgendb.query_engine import QueryEngine, QueryResults
from matgendb.tests import common

has_mongo = common.has_mongo()

class SandboxTest(unittest.TestCase):

    SBX = "testing"
    N = 100

    def qtx(self, crit, props):
        if props==None:
            props = {}
        crit['sbxd.e_above_hull'] = crit['e_above_hull']
        props['sbxd'] = {"$elemMatch": { "id": self.SBX } }
        del crit['e_above_hull']

    def rtx(self, doc):
        doc['add_fake_field'] = 'test value'
        for item in doc['sbxd']:
            if item['id'] == self.SBX:
                doc['e_above_hull'] = item['e_above_hull']
        return doc

    def setUp(self):
        # Try a real mongodb
        if has_mongo:
            self.conn = pymongo.MongoClient()
            self.db_name = 'test'
            self.db = self.conn[self.db_name]
            self.coll_name = 'sandboxes_{}'.format(uuid.uuid4())
            self.coll = self.db[self.coll_name]
            for i in range(self.N):
                core_v, sbx_v = 0.1 * (i + 1), -0.1 * (i + 1)
                doc = {'task_id': 'mp-{:d}'.format(1000 + i),
                       'sbxd': [{'id': 'core', 'e_above_hull': core_v},
                                {'id': self.SBX, 'e_above_hull': sbx_v}],
                       'sbxn': ['core', self.SBX]}
                doc.update({'state': 'successful'})
                if i < 2:
                    pprint.pprint(doc)
                self.coll.insert(doc)

    def tearDown(self):
        if has_mongo:
            self.db.drop_collection(self.coll_name)



    @unittest.skipUnless(has_mongo, 'requires MongoDB server')
    def test_no_post_funcs(self):
        qe = QueryEngine(connection=self.conn,
                         database=self.db_name,
                         collection=self.coll_name,
                         aliases={},
                         query_post=[], result_post=[])
        cursor = qe.query()
        self.assertTrue(isinstance(cursor, QueryResults))
        n = 0
        for rec in cursor:
            pprint.pprint("RESULT: {}".format(rec))
            # No Post proccessing should be done
            self.assertTrue('e_above_hull' not in rec)
            self.assertTrue('add_fake_field' not in rec)

            self.assertTrue('sbxd' in rec)
            n += 1
        # should find all tasks
        self.assertEqual(n, self.N)



    @unittest.skipUnless(has_mongo, 'requires MongoDB server')
    def test_mongo_find(self):

        # 
        qe = QueryEngine(connection=self.conn,
                         database=self.db_name,
                         collection=self.coll_name,
                         aliases={},
                         query_post=[self.qtx], result_post=[self.rtx])
        result = self._test_find(qe, 
                                 criteria={'e_above_hull': {'$lte': 0.0}},
                                 properties={})

    @unittest.skipUnless(has_mongo, 'requires MongoDB server')
    def test_with_properties(self):

        # 
        qe = QueryEngine(connection=self.conn,
                         database=self.db_name,
                         collection=self.coll_name,
                         aliases={},
                         query_post=[self.qtx], result_post=[self.rtx])
        result = self._test_find(qe, 
                                 criteria={'e_above_hull': {'$lte': 0.0}},
                                 properties=['e_above_hull', 'sbxd', 'add_fake_field'])


    def _test_find(self, qe, properties, criteria):
        cursor = qe.query(properties=properties,
                          criteria=criteria)
        self.assertTrue(isinstance(cursor, QueryResults))
        n = 0
        for rec in cursor:
            pprint.pprint("RESULT: {}".format(rec))
            self.assertTrue(rec['e_above_hull'] < 0)
            self.assertEqual(rec['add_fake_field'], 'test value')
            n += 1
        # should find all tasks
        self.assertEqual(n, self.N)


    @unittest.skipUnless(has_mongo, 'requires MongoDB server')
    def test_queryresult(self):

        qe = QueryEngine(connection=self.conn,
                         database=self.db_name,
                         collection=self.coll_name,
                         aliases={},
                         query_post=[self.qtx], result_post=[self.rtx])
        result = qe.query(criteria={'e_above_hull': {'$lte': 0.0}}).sort('sbxd.e_above_hull', pymongo.ASCENDING)
        self.assertTrue(isinstance(result, QueryResults))
        self.assertEqual(len(result), self.N)
        self.assertTrue(result[0]['e_above_hull'] < 0)

if __name__ == '__main__':
    unittest.main()


"""
Diff collections, as sets
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '3/29/13'

# System
import logging
import re
import time
# Package
from matgendb import util
from matgendb.query_engine import QueryEngine
from matgendb.dbconfig import normalize_auth

_log = logging.getLogger("mg.vv.diff")


class IID:
    _value = 0

    @classmethod
    def next(cls):
        cls._value += 1
        return cls._value


class Differ:
    """Calculate difference between two collections, based solely on a
    selected key.

    As noted in :func:`diff`, this will not work with huge datasets, as it stores
    all the keys in memory in order to do a "set difference" using Python sets.
    """

    #: Keys in result dictionary.
    MISSING, NEW, CHANGED = 'missing', 'additional', 'different'

    #: CHANGED result fields
    CHANGED_MATCH_KEY = "match type"
    CHANGED_MATCH_DELTA = "delta"
    CHANGED_MATCH_EXACT = "exact"
    CHANGED_OLD = "old"
    CHANGED_NEW = "new"
    CHANGED_DELTA = "delta"

    #: for missing property
    NO_PROPERTY = "__MISSING__"

    def __init__(self, key='_id', props=None, info=None, fltr=None, deltas=None):
        """Constructor.

        :param key: Field to use for identifying records
        :param props: List of fields to use for matching records
        :param info: List of extra fields to retrieve from (and show) for each record.
        :param fltr: Filter for records, a MongoDB query expression
        :param deltas: {prop: delta} to check. 'prop' is a string, 'delta' is an instance of :class:`Delta`.
                       Any key for 'prop' not in parameter 'props' will get added.
        :type deltas: dict
        :raise: ValueError if some delta does not parse.
        """
        self._key_field = key
        self._props = [] if props is None else props
        self._info = [] if info is None else info
        self._filter = fltr if fltr else {}
        self._prop_deltas = {} if deltas is None else deltas
        self._all_props = list(set(self._props[:] +
                                   list(self._prop_deltas.keys())))

    def diff(self, c1, c2, only_missing=False, only_values=False, allow_dup=False):
        """Perform a difference between the 2 collections.
        The first collection is treated as the previous one, and the second
        is treated as the new one.

        Note: this is not 'big data'-ready; we assume all the records can fit in memory.

        :param c1: Collection (1) config file, or QueryEngine object
        :type c1: str or QueryEngine
        :param c2: Collection (2) config file, or QueryEngine object
        :type c2: str or QueryEngine
        :param only_missing: Only find and return self.MISSING; ignore 'new' keys
        :param only_values: Only find and return self.CHANGED; ignore new or missing keys
        :param allow_dup: Allow duplicate keys, otherwise fail with ValueError
        :return: dict with keys self.MISSING, self.NEW (unless only_missing is True), & self.CHANGED,
                 each a list of records with the key and
                 any other fields given to the constructor 'info' argument.
                 The meaning is: 'missing' are keys that are in c1 not found in c2
                 'new' is keys found in c2 that are not found in c1, and 'changed' are records
                 with the same key that have different 'props' values.
        """
        # Connect.
        _log.info("connect.start")
        if isinstance(c1, QueryEngine):
            engines = [c1, c2]
        else:
            engines = []
            for cfg in c1, c2:
                settings = util.get_settings(cfg)
                if not normalize_auth(settings):
                    _log.warn("Config file {} does not have a username/password".format(cfg))
                settings["aliases_config"] = {"aliases": {}, "defaults": {}}
                engine = QueryEngine(**settings)
                engines.append(engine)
        _log.info("connect.end")

        # Query DB.
        keys = [set(), set()]
        eqprops = [{}, {}]
        numprops = [{}, {}]

        # Build query fields.
        fields = dict.fromkeys(self._info + self._all_props + [self._key_field], True)
        if not '_id' in fields:  # explicitly remove _id if not given
            fields['_id'] = False

        # Initialize for query loop.
        info = {}  # per-key information
        has_info, has_props = bool(self._info), bool(self._all_props)
        has_numprops, has_eqprops = bool(self._prop_deltas), bool(self._props)
        _log.info("query.start query={} fields={}".format(self._filter, fields))
        t0 = time.time()

        # Main query loop.
        for i, coll in enumerate(engines):
            _log.debug("collection {:d}".format(i))
            count, missing_props = 0, 0
            for rec in coll.query(criteria=self._filter, properties=fields):
                count += 1
                # Extract key from record.
                try:
                    key = rec[self._key_field]
                except KeyError:
                    _log.critical("Key '{}' not found in record: {}. Abort.".format(
                        self._key_field, rec))
                    return {}
                if not allow_dup and key in keys[i]:
                    raise ValueError("Duplicate key: {}".format(key))
                keys[i].add(key)
                # Extract numeric properties.
                if has_numprops:
                    pvals = {}
                    for pkey in self._prop_deltas.keys():
                        try:
                            pvals[pkey] = float(rec[pkey])
                        except KeyError:
                            #print("@@ missing {} on {}".format(pkey, rec))
                            missing_props += 1
                            continue
                        except (TypeError, ValueError):
                            raise ValueError("Not a number: collection={c} key={k} {p}='{v}'"
                                             .format(k=key, c=("old", "new")[i], p=pkey, v=rec[pkey]))
                    numprops[i][key] = pvals
                # Extract properties for exact match.
                if has_eqprops:
                    try:
                        propval = tuple([(p, str(rec[p])) for p in self._props])
                    except KeyError:
                        missing_props += 1
                        #print("@@ missing {} on {}".format(pkey, rec))
                        continue
                    eqprops[i][key] = propval

                # Extract informational fields.
                if has_info:
                    if key not in info:
                        info[key] = {}
                    for k in self._info:
                        info[key][k] = rec[k]

            # Stop if we don't have properties on any record at all
            if 0 < count == missing_props:
                _log.critical("Missing one or more properties on all {:d} records"
                              .format(count))
                return {}
            # ..but only issue a warning for partially missing properties.
            elif missing_props > 0:
                _log.warn("Missing one or more properties for {:d}/{:d} records"
                          .format(missing_props, count))
        t1 = time.time()
        _log.info("query.end sec={:f}".format(t1 - t0))

        # Compute missing and new keys.
        if only_values:
            missing, new = [], []
        else:
            _log.debug("compute_difference.start")
            missing, new = keys[0] - keys[1], []
            if not only_missing:
                new = keys[1] - keys[0]
            _log.debug("compute_difference.end")

        # Compute mis-matched properties.
        if has_props:
            changed = self._changed_props(keys, eqprops, numprops, info,
                                          has_eqprops=has_eqprops, has_numprops=has_numprops)
        else:
            changed = []

        # Build result.
        _log.debug("build_result.begin")
        result = {}
        if not only_values:
            result[self.MISSING] = []
            for key in missing:
                rec = {self._key_field: key}
                if has_info:
                    rec.update(info.get(key, {}))
                result[self.MISSING].append(rec)
            if not only_missing:
                result[self.NEW] = []
                for key in new:
                    rec = {self._key_field: key}
                    if has_info:
                        rec.update(info.get(key, {}))
                    result[self.NEW].append(rec)
        result[self.CHANGED] = changed
        _log.debug("build_result.end")

        return result

    def _changed_props(self, keys=None, eqprops=None, numprops=None, info=None,
                       has_numprops=False, has_eqprops=False):
        changed = []
        _up = lambda d, v: d.update(v) or d   # functional dict.update()
        for key in keys[0].intersection(keys[1]):
            # Numeric property comparisons.
            if has_numprops:
                for pkey in self._prop_deltas:
                    oldval, newval = numprops[0][key][pkey], numprops[1][key][pkey]
                    if self._prop_deltas[pkey].cmp(oldval, newval):
                        change = {self.CHANGED_MATCH_KEY: self.CHANGED_MATCH_DELTA, self._key_field: key, "property": pkey,
                                  self.CHANGED_OLD: "{:f}".format(oldval), self.CHANGED_NEW: "{:f}".format(newval),
                                  "rule": self._prop_deltas[pkey],
                                  self.CHANGED_DELTA: "{:f}".format(newval - oldval)}
                        changed.append(_up(change, info[key]) if info else change)
            # Exact property comparison.
            if has_eqprops:
                if not eqprops[0][key] == eqprops[1][key]:
                    change = {self.CHANGED_MATCH_KEY: self.CHANGED_MATCH_EXACT, self._key_field: key,
                              self.CHANGED_OLD: eqprops[0][key], self.CHANGED_NEW: eqprops[1][key]}
                    changed.append(_up(change, info[key]) if info else change)
        return changed


class Delta:
    """Delta between two properties.

    Syntax:
        +-       Change in sign, 0 not included
        +-=      Change in sign, + to 0 or - to 0 included
        +-X      abs(new - old) > X
        +X-Y     (new - old) > X or (old - new) > Y
        +-X=     abs(new - old) >= X
        +X-Y=    (new - old) >= X or (old - new) >= Y
        +X[=]   Just look in '+' direction
        -Y[=]   Just look in '-' direction
        ...%     Instead of (v2 - v1), use 100*(v2 - v1)/v1
    """
    _num = "\d+(\.\d+)?"
    _expr = re.compile("(?:"
                       "\+(?P<X>{n})?-(?P<Y>{n})?|"  # both + and -
                       "\+(?P<X2>{n})?|"              # only +
                       "-(?P<Y2>{n})?"                # only -
                       ")"
                       "(?P<eq>=)?(?P<pct>%)?".format(n=_num))

    def __init__(self, s):
        """Constructor.

        :param s: Expression string
        :type s: str
        :raises: ValueError if it doesn't match the syntax
        """
        # Match expression.
        m = self._expr.match(s)
        if m is None:
            raise ValueError("Bad syntax for delta '{}'".format(s))
        if m.span()[1] != len(s):
            p = m.span()[1]
            raise ValueError("Junk at end of delta '{}': {}".format(s, s[p:]))

        # Save a copy of orig.
        self._orig_expr = s

        # Initialize parsed values.
        self._sign = False
        self._dx, self._dy = 0, 0
        self._pct = False           # %change
        self._eq = False            # >=,<= instead of >, <

        # Set parsed values.
        d = m.groupdict()
        #print("@@ expr :: {}".format(d))
        if all((d[k] is None for k in ('X', 'Y', 'X2', 'Y2'))):
            # Change in sign only
            self._sign = True
            self._eq = d['eq'] is not None
        elif d['X'] is not None and d['Y'] is None:
            raise ValueError("Missing value for negative delta '{}'".format(s))
        else:
            if d['X2'] is not None:
                # Positive only
                self._dx = float(d['X2'])
                self._dy = None
            elif d['Y2'] is not None:
                # Negative only
                self._dx = None
                self._dy = -float(d['Y2'])
            else:
                # Both
                self._dy = -float(d['Y'])
                self._dx = float(d['X'] or d['Y'])
            self._eq = d['eq'] is not None
            self._pct = d['pct'] is not None
            #print("@@ dx,dy eq,pct = {},{}  {},{}".format(self._dx, self._dy, self._eq, self._pct))

        # Pre-calculate comparison function.
        if self._sign:
            self._cmp = self._cmp_sign
        elif self._pct:
            self._cmp = self._cmp_val_pct
        else:
            self._cmp = self._cmp_val_abs

        self._json_id = None  # for repeated serialization

    def __str__(self):
        return self._orig_expr

    def as_json(self):
        if self._json_id:
            # only serialize fully the first time
            return {'delta': {'id': self._json_id}}
        dtype = 'abs' if self._eq else 'pct'
        incl = self._eq
        self._json_id = next(IID)
        return {
            'delta': {
                'plus': self._dx,
                'minus': self._dy,
                'type': dtype,
                'endpoints': incl,
                'id': self._json_id
            }
        }

    def cmp(self, old, new):
        """Compare numeric values with delta expression.

        Returns True if delta matches (is as large or larger than) this class' expression.

        Delta is computed as (new - old).

        :param old: Old value
        :type old: float
        :param new: New value
        :type new: float
        :return: True if delta between old and new is large enough, False otherwise
        :rtype: bool
        """
        return self._cmp(old, new)

    def _cmp_sign(self, a, b):
        if self._eq:
            return (a < 0 <= b) or (a > 0 >= b)
        return (a < 0 < b) or (a > 0 > b)

    def _cmp_val_abs(self, a, b):
        return self._cmp_val(b - a)

    def _cmp_val_pct(self, a, b):
        if a == 0:
            return False
        return self._cmp_val(100.0 * (b - a) / a)

    def _cmp_val(self, delta):
        oor = False  # oor = out-of-range
        if self._eq:
            if self._dx is not None:
                oor |= delta >= self._dx
            if self._dy is not None:
                oor |= delta <= self._dy
        else:
            if self._dx is not None:
                oor |= delta > self._dx
            if self._dy is not None:
                oor |= delta < self._dy
        return oor



"""
Utility module for schema validation, for builder testing
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '11/1/13'

import datetime
import glob
import json
import os
import re
import time

# Default schema version
DEFAULT_VERSION = '1.0.0'

# Indicates any version
ANY_VERSION = "*.*.*"

# denotes optional field/scalar
OPTIONAL_FLAG = '?'

# marker of specialness
SPECIAL = '__'
SPECIAL_LEN = len(SPECIAL)

# Regex for values
# Note that the special prefix/suffix is optional
VALUE_RE = re.compile("(?:{spec})?([a-zA-Z]+)(?:{spec})?\s*(.*)"
                      .format(spec=SPECIAL))

# Global obj with all collected schemas
schemata = {}


class SchemaError(Exception):
    """Base class of all errors raised by schema creation or validation.
    """
    pass


class SchemaTypeError(SchemaError):
    def __init__(self, typename):
        SchemaError.__init__(self, "bad type ({})".format(typename))


class SchemaPathError(SchemaError):
    pass


class SchemaParseError(SchemaError):
    pass


def add_schemas(path, ext="json"):
    """Add schemas from files in 'path'.

    :param path: Path with schema files. Schemas are named by their file,
                 with the extension stripped. e.g., if path is "/tmp/foo",
                 then the schema in "/tmp/foo/bar.json" will be named "bar".
    :type path: str
    :param ext: File extension that identifies schema files
    :type ext: str
    :return: None
    :raise: SchemaPathError, if no such path. SchemaParseError, if a schema
            is not valid JSON.
    """
    if not os.path.exists(path):
        raise SchemaPathError()
    filepat = "*." + ext if ext else "*"
    for f in glob.glob(os.path.join(path, filepat)):
        with open(f, 'r') as fp:
            try:
                schema = json.load(fp)
            except ValueError:
                raise SchemaParseError("error parsing '{}'".format(f))
        name = os.path.splitext(os.path.basename(f))[0]
        schemata[name] = Schema(schema)


def get_schema(name):
    """Get schema by name

    :param name: name of schema
    :type name: str
    :return: new schema
    :rtype: Schema
    :raise: KeyError, if schema is not found
    """
    return schemata[name]


def load_schema(file_or_fp):
    """Load schema from file.

    :param file_or_fp: File name or file object
    :type file_or_fp: str, file
    :raise: IOError if file cannot be opened or read, ValueError if
            file is not valid JSON or JSON is not a valid schema.
    """
    fp = open(file_or_fp, 'r') if isinstance(file_or_fp, str) else file_or_fp
    obj = json.load(fp)
    schema = Schema(obj)
    return schema

## Validator classes


class HasMeta:
    """Mix-in class to handle metadata.
    Adds the 'meta' class attribute.
    """
    FIELD_SEP = ","
    KV_SEP = ":"

    def __init__(self, meta):
        """Create with new metadata (which may be empty).
        The metadata should be dict-like (i.e. __getitem__ and __setitem__) or
        be a string in the format "name1: value1, name2: value2, ...", where the
        white space is optional.

        :param meta: Init with this metadata.
        :type meta: str or dict
        """
        if not meta:
            self.meta = {}
        elif isinstance(meta, basestring):
            # build a dict from key-value pairs
            parts = meta.split(self.FIELD_SEP)
            self.meta = {k: v for k, v in map(lambda fld: fld.split(':', 1), parts)}
        else:
            # assume dict-like
            self.meta = meta

    def add_meta(self, key, value):
        self.meta[key] = value


class Schema(HasMeta):

    # enum for navigating hierarchies
    IS_LIST, IS_DICT, IS_SCALAR = 0, 1, 2

    def __init__(self, schema, optional=False, meta=''):
        HasMeta.__init__(self, meta)
        self.is_optional = optional
        self._schema = self._parse(schema)
        self._json_schema = None
        self._json_schema_keys = {}

    def validate(self, doc, path="(root)"):
        t = self._whatis(doc)
        if t != self._type:
            return self._vresult(path, "type mismatch: {} != {}", self._typestr(t), self)
        if t == self.IS_LIST:
            if len(doc) == 0:
                return None
            return self._schema[0].validate(doc[0], path=path + "[0]")
        elif t == self.IS_DICT:
            # fail if document is missing any required keys
            dkeys = set(doc.keys())
            skeys = set(filter(lambda k: not self._schema[k].is_optional,
                               self._schema.keys()))
            if skeys - dkeys:
                return self._vresult(path, "missing keys: ({})".format(', '.join(skeys - dkeys)))
            # check each item in document
            for k, v in doc.items():
                if k in self._schema:
                    result = self._schema[k].validate(doc[k], path=path + "." + k)
                    if result is not None:
                        return result
                else:
                    pass  # do nothing for 'extra' keys
                    #return self._vresult(path, "missing key: {}", k)
        else:
            if not self._schema.check(doc):
                return self._vresult(path, "bad value '{}' for type {}",
                                     doc, self._schema)

    def json_schema(self, **add_keys):
        """Convert our compact schema representation to the standard, but more verbose,
        JSON Schema standard.

        Example JSON schema: http://json-schema.org/examples.html
        Core standard: http://json-schema.org/latest/json-schema-core.html

        :param add_keys: Key, default value pairs to add in,
                         e.g. description=""
        """
        self._json_schema_keys = add_keys
        if self._json_schema is None:
            self._json_schema = self._build_schema(self._schema)
        return self._json_schema

    def _build_schema(self, s):
        """Recursive schema builder, called by `json_schema`.
        """
        w = self._whatis(s)
        if w == self.IS_LIST:
            w0 = self._whatis(s[0])
            js = {"type": "array",
                  "items": {"type": self._jstype(w0, s[0])}}
        elif w == self.IS_DICT:
            js = {"type": "object",
                  "properties": {key: self._build_schema(val) for key, val in s.items()}}
            req = [key for key, val in s.items() if not val.is_optional]
            if req:
                js["required"] = req
        else:
            js = {"type": self._jstype(w, s)}
        for k, v in self._json_schema_keys.items():
            if k not in js:
                js[k] = v
        return js

    def _jstype(self, stype, sval):
        """Get JavaScript name for given data type, called by `_build_schema`.
        """
        if stype == self.IS_LIST:
            return "array"
        if stype == self.IS_DICT:
            return "object"
        if isinstance(sval, Scalar):
            return sval.jstype
        # it is a Schema, so return type of contents
        v = sval._schema
        return self._jstype(self._whatis(v), v)

    def _vresult(self, path, fmt, *args):
        meta_info = ''
        if self.meta and 'desc' in self.meta:
            meta_info = '="{}"'.format(self.meta['desc'])
        return "{}{}: ".format(path, meta_info) + fmt.format(*args)

    def _parse(self, value):
        t = self._type = self._whatis(value)
        if t == self.IS_LIST:
            return [Schema(value[0])]
        elif t == self.IS_DICT:
            r = {}
            for k, v in value.items():
                # skip metadata keys
                if k.startswith(SPECIAL) and k.endswith(SPECIAL):
                    self.add_meta(k[SPECIAL_LEN:-SPECIAL_LEN], v)
                    continue
                # look for optional flag at start of key
                opt_flag = self.is_optional
                if k[0] == OPTIONAL_FLAG:
                    k = k[1:]
                    opt_flag = True
                elif k in ('@class', '@module'):  # optionalize the cruft
                    opt_flag = True
                # parse value and assign
                r[k] = Schema(v, optional=opt_flag)
            return r
        else:
            optional = self.is_optional
            if len(value) and value[0] == OPTIONAL_FLAG:
                value = value[1:]
                optional = True
            vinfo = VALUE_RE.match(value)
            if not vinfo:
                raise ValueError("bad type format, must be __<type>__ got {}"
                                 .format(value))
            dtype, meta = vinfo.groups()
            return Scalar(dtype, optional=optional, meta=meta)

    def _whatis(self, obj):
        if isinstance(obj, list):
            return self.IS_LIST
        elif isinstance(obj, dict):
            return self.IS_DICT
        return self.IS_SCALAR

    def _typestr(self, t):
        return ("list", "dict", "scalar")[t]

    def __str__(self):
        return self._typestr(self._type)

    def __repr__(self):
        return "document::{}".format(self)


def _is_datetime(d):
    return isinstance(d, datetime.datetime) or \
        isinstance(d, time.struct_time)


class Scalar(HasMeta):
    # For each typecode, a function that returns True/False on whether
    # its argument matches.

    TYPES = {
        'string': lambda x: isinstance(x, basestring),
        'bool': lambda x: x is True or x is False,
        'datetime': _is_datetime,
        'date': _is_datetime,
        'float': lambda x: isinstance(x, float),
        'int': lambda x: isinstance(x, int),
        'null': lambda x: x is None,
        'array': lambda x: isinstance(x, list),
        'object': lambda x: isinstance(x, dict)
    }

    def __init__(self, typecode, optional=False, meta=''):
        self.is_optional = optional
        self._type = typecode
        HasMeta.__init__(self, meta)
        try:
            self.check = self.TYPES[typecode]
        except KeyError:
            raise SchemaTypeError(typecode)

    JSTYPES = {
        "datetime": "string", "date": "string",
        "string": "string",
        "bool": "boolean",
        "int": "integer",
        "float": "number",
        "null": "null"
    }

    @property
    def jstype(self):
        """Return JavaScript type.
        """
        return self.JSTYPES[self._type]

    def __str__(self):
        return self._type

    def __repr__(self):
        return "scalar::{}".format(self)




# -*- coding: utf8 -*-
"""
Run validation-related actions against (MongoDB) databases.
Database connections are configured from a file
and the constraints to validate against can come from a file or the command-line.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '3/29/13'

import argparse
import logging
import os
import pymongo
import smtplib
import socket
import sys
import time
import traceback
import yaml
import importlib
# local module
from matgendb.util import get_settings, get_collection
from matgendb.vv.validate import ConstraintSpec, Validator
from matgendb.vv.validate import ValidatorSyntaxError, DBError
from matgendb.vv import report
from matgendb.vv.util import Timing, ElapsedTime, letter_num
from matgendb.vv.util import YamlConfig, args_kvp_nodup, args_list
from matgendb.vv import diff
#from matgendb.vv import mquery
from smoqe.query import to_mongo, BadExpression


# Logging
# -------

_log_name = 'mg'
_log = logging.getLogger(_log_name)
_handler = logging.StreamHandler()
_handler.setFormatter(logging.Formatter('%(levelname)-7s [%(name)s] %(message)s'))
_log.addHandler(_handler)


# Constants
# ---------

DB_KEY = '_db'              # in YAML file
ALIASES_KEY = '_aliases'    # in YAML file

PATTERN_KEY_PREFIX_IGNORE = '_'


# Exceptions
# ----------


class FunctionError(Exception):
    """Error with a top-level function.
    """
    def __init__(self, command, why):
        msg = "In '{}' command: {}".format(command, why)
        Exception.__init__(self, msg)


class ArgumentError(Exception):
    """Command-line arg. error
    """

    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg

class InputSyntaxError(Exception):
    """Error with syntax of some input file.
    """
    def __init__(self, target, desc):
        msg = 'Syntax error in "{}": {}'.format(target, desc)
        Exception.__init__(self, msg)

# Utility functions
# -----------------


def wstrip(s):
    return s.strip()


def textlist(items):
    return ', '.join(items)


def mapdump(d):
    return textlist(('{}={}'.format(k, v) for k, v in d.iteritems()))


def nvpstr(s):
    """Return tuple (name, value) from string in form 'name=value'.
    Used as type for argparse.

    :raise: ValueError if input string is invalid
    """
    p = s.find('=')
    if p < 0:
        raise ValueError('expected "name=value", got "{}"'.format(s))
    nvp = (s[:p].strip(), s[p + 1:].strip())
    #print("NVP: {}".format(nvp))
    return nvp


def _parse_email(s):
    """Parse an email specification into component parts.

    Keys:
        - sender
        - recipients
        - host
        - port

    :return: Parsed email information
    :rtype: dict
    """
    d = {}
    # parse command-line args
    items = s.split(':')
    if len(items) < 2:
        raise ValueError('from:to required in "{}"'.format(s))
    if len(items) > 4:
        raise ValueError('too many values in "{}"'.format(s))
    d['sender'] = items[0]
    d['recipients'] = [x.strip() for x in items[1].split(',')]
    if len(items) > 2:
        d['server'] = items[2]
    if len(items) > 3:
        if '/' in items[3]:
            p1, p2 = items[3].split('/')
            d['port'], d['subject'] = int(p1), p2
        else:
            d['port'] = int(items[3])
    return d


def _config_email(fname):
    """Get an email spec. from a config file.

    :return: Parsed email information
    :rtype: dict
    """
    d = {}
    try:
        f = open(fname)
    except OSError:
        raise ValueError('cannot read from file "{}"'.format(fname))
    y = yaml.safe_load(f)
    info = y.get('_email', None)
    if info is None:
        return None
    try:
        d['sender'] = info['from']
        d['recipients'] = info['to']
        if isinstance(d['recipients'], str):
            d['recipients'] = [d['recipients']]
    except KeyError:
        raise ValueError('from/to not found in file {} mapping key "_email"'
                         .format(fname))
    if 'port' in info:
        d['port'] = int(info['port'])
    if 'host' in info:
        d['host'] = info['host']
    if 'subject' in info:
        d['subject'] = info['subject']
    return d


def _splitup(s, sep, n1, n2):
    parts = s.split(sep)
    n = len(parts)
    if not (n1 <= n <= n2):
        raise ValueError("Wrong number of '{}'-separated parts: {:d}".format(sep, n))
    return parts


def parse_email_args(msg, server):
    mp = _splitup(msg, '/', 2, 3)
    sp = _splitup(server, ':', 1, 2)
    return {"sender": mp[0],
            "recipients": [x.strip() for x in mp[1].split(',')],
            "subject": "" if len(mp) < 3 else mp[2],
            "server": sp[0],
            "port": int(sp[1]) if len(sp) > 1 else smtplib.SMTP_PORT}


def _dbinfo(cfg):
    return "{host}/{collection}.{database}".format(**cfg)


# Subcommands
# -----------


def command_diff(args):
    """Sub-command to calculate the difference of 2 collections.
    """
    # Check that conf files exist
    try:
        map(open, (args.old, args.new))
    except IOError as err:
        raise FunctionError("diff", "Cannot open configuration file: {}".format(err))

    # Parse query/filter, if provided
    fltr = {}
    if args.fltr is not None:
        try:
            fltr = to_mongo(args.fltr)
        except BadExpression as err:
            raise FunctionError("diff", "Bad query filter: {}".format(err.details))
        _log.info("Filter (original => mongo): {} => {}".format(args.fltr, fltr))

    # Parse numeric change fields, if given
    deltas = {}
    if args.numprops:
        for field, expr in args.numprops.iteritems():
            try:
                deltas[field] = diff.Delta(expr)
            except ValueError as err:
                raise FunctionError("diff", "Bad numeric expr for field '{}': {}"
                                    .format(field, err))

    # Instantiate main class.
    df = diff.Differ(key=args.key, info=args.info, props=args.properties,
                     fltr=fltr, deltas=deltas)

    # Run diff.
    t0 = time.time()
    try:
        r = df.diff(args.old, args.new, only_missing=args.missonly,
                    only_values=args.changeonly)
    except Exception as err:
        if _log.getEffectiveLevel() in (logging.DEBUG,):
            exc_str = traceback.format_exc()
            err = "{}\nTraceback:\n{}".format(err, exc_str)
        raise FunctionError("diff", "Runtime error: {}".format(err))
    t1 = time.time()

    # Build report.
    stime = lambda t: time.strftime("%Y-%m-%d %H:%M", time.localtime(t))
    meta = {'start_time': stime(t0), 'end_time': stime(t1),
            'elapsed': t1 - t0,
            'filter': fltr,
            'db1': _dbinfo(get_settings(args.old)),
            'db2': _dbinfo(get_settings(args.new)),
            'args': {k: v for k, v in vars(args).iteritems()
                     if not k.startswith("func")}
    }
    meta.update(system_metadata())
    fmt_kwargs = dict(key=args.key)
    if args.format == "html":
        fmt_kwargs['url'] = args.rest_url
        fmt_kwargs['email_mode'] = args.rpt_email

    make_report = lambda f: getattr(report, "Diff{}Formatter".format(f.title()))(meta, **fmt_kwargs).format(r)

    # Email, print, and/or db-insert a report
    # (a) email
    if args.rpt_email:
        kw = parse_email_args(args.rpt_email, args.email_server)
        emailer = report.Emailer(**kw)
        fmt = args.format or "html"
        text = make_report(fmt)
        emailer.send(text, ("text/plain", "text/html")[fmt == "html"])
    # (b) print
    if args.rpt_print or (not args.rpt_email and not args.rpt_print and not args.rpt_db):
        if args.format == "json":
            fmt_kwargs['pretty'] = True
        text = make_report(args.format or "text")
        print(text)
    # (c) DB
    if args.rpt_db:
        try:
            coll = get_collection(args.rpt_db, admin=True)
        except Exception as err:
            raise FunctionError("diff", "Connecting to report DB: {}".format(err))
        try:
            #coll.database.add_son_manipulator(report.DiffJsonFormatter.Manipulator())
            doc = report.DiffJsonFormatter(meta).document(r)
            coll.insert(doc)
        except pymongo.errors.OperationFailure as err:
            raise FunctionError("diff", "Inserting to report DB: {}".format(err))

    return 0


def system_metadata():
    """Return some simple system metadata.
    """
    try:
        user_name = os.getlogin()
    except OSError:
        user_name = 'unknown'
    host_name = socket.gethostname()
    cmdline = " ".join(sys.argv)
    return {'user': user_name,
            'host': host_name,
            "cmdline": cmdline}


def command_validate(args, formatters):
    """Run validation command.
    """
    retcode = 0  # return code
    # Connect and authenticate
    fname = args.config_file if args.config_file else "db.json"
    try:
        config = get_settings(args.config_file)
    except ValueError as err:
        raise ArgumentError('Cannot parse configuration "{}": {}'.format(fname, err))
    try:
        conn = pymongo.MongoClient(config['host'], config['port'])
    except pymongo.errors.ConnectionFailure as err:
        raise ArgumentError('Cannot connect to server: {}'.format(err))
    db_key = 'database'
    if not db_key in config:
        raise ArgumentError('Configuration in "{}" is missing "{}"'.format(fname, db_key))
    database_name = config[db_key]
    db = conn[database_name]
    user, passwd = None, None
    # try both readonly and admin credentials
    for utype in 'readonly', 'admin':
        if (utype + '_user') in config:
            try:
                user, passwd = config[utype + '_user'], config[utype + '_password']
            except KeyError:
                raise ArgumentError('Configuration error in file {0}: {1}_user but no {1}_password'
                                    .format(args.config_file, utype))
    if user and passwd:
        if not db.authenticate(user, passwd):
            raise ArgumentError('Cannot authenticate to database as user {}'.format(user))
    else:
        _log.info('Connecting to DB "{}" without authentication'.format(config[db_key]))
        # get collection name if given
    if 'collection' in config:
        db_coll_name = config['collection']
    else:
        db_coll_name = None

    # Init constraints from file or command-line
    constraints_from_file = args.constfile is not None
    if constraints_from_file:
        try:
            f = open(args.constfile, 'rb')
        except IOError:
            if not os.path.exists(args.constfile):
                raise ArgumentError('constraints file "{}" not found'.format(args.constfile))
            else:
                raise ArgumentError('constraints file "{}" found, but unreadable'.format(args.constfile))
        with Timing('load.yaml', file=f.name):
            constraints = yaml.safe_load(f)
    else:
        coll_name = args.coll
        if coll_name is None:
            if db_coll_name is None:
                raise ArgumentError('collection name not found')
            else:
                coll_name = db_coll_name
        if not args.constraint:
            raise ArgumentError('at least one constraint is required')
            # re-split command-line arguments on commas (not whitespace)
        plist = ' '.join(args.constraint).split(',')
        # make dict(key=collection) of list of the constraints, just like YAML config file
        constraints = {coll_name: [plist]}
        _log.debug("command-line constraints dump: {}".format(constraints))

    # Init aliases from file or command-line
    aliases = {}
    if constraints_from_file:
        astr = ""
        try:
            astr = constraints.get(ALIASES_KEY, {})
            aliases = dict(map(nvpstr, astr))
        except ValueError as err:
            raise ArgumentError('aliases should be list of name=value, got "{}": {}'.format(astr, err))
        except AttributeError:
            if constraints is None:
                raise ArgumentError('constraints file "{}" is empty'.format(args.constfile))
    elif len(args.aliases) > 0:
        aliases = dict(args.aliases)

    # Init email from file or command-line
    emailer, email_kw = None, None
    if args.email is not None:
        email_kw = _parse_email(args.email)
    elif args.constfile:
        email_kw = _config_email(args.constfile)
    if email_kw is not None:
        _kwd = {'subject': 'MP Validation Report'}
        _kwd.update(email_kw)
        _log.info("Will email report to: {}".format(mapdump(_kwd)))
        emailer = report.Emailer(**_kwd)

    # Initialize report
    sys_meta = system_metadata()
    rpt_hdr = report.ReportHeader("Materials Project Validation Report")
    rpt_hdr.add('Report time', time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime()))
    rpt_hdr.add('Report user', sys_meta['user'])
    rpt_hdr.add('Report host', sys_meta['host'])
    rpt_hdr.add('Database', database_name)
    rpt_hdr.add('Limit', args.limit)
    rpt = report.Report(rpt_hdr)
    # pick report formatter
    fmt = args.report_format.lower()
    formatter_class = formatters.get(fmt, None)
    if formatter_class is None:
        raise ArgumentError('Unknown format "{}" for --format, choose from: {}'
                            .format(fmt, textlist(formatters.keys())))

    # Run validation for each collection
    validator = Validator(aliases=aliases, max_violations=args.limit,
                              max_dberrors=10, add_exists=args.must_exist)
    if args.progress > 0:
        validator.set_progress(args.progress)
    elapsed = ElapsedTime()
    with Timing("validate", log=_log, elapsed=elapsed):
        for coll_name, constraint_spec_cfg in constraints.iteritems():
            _log.debug("validate {}".format(coll_name))
            sect_hdr = report.SectionHeader(title='Collection "{}"'.format(coll_name))
            rpt_sect = report.ReportSection(sect_hdr)
            if coll_name.startswith(PATTERN_KEY_PREFIX_IGNORE):
                continue
            coll = db[coll_name]
            try:
                try:
                    cspec = ConstraintSpec(constraint_spec_cfg)
                except ValueError as err:
                    _log.error('processing constraints for {}: {}'.format(coll_name, err))
                    break
                    #sect = get_constraint_sections(constraint_spec)
                vsect = 0
                try:
                    vresult = validator.validate(coll, cspec, subject=coll_name)
                except ValidatorSyntaxError as err:
                    target = 'Collection = {}, Constraint section = {}'.format(
                        coll.name, err)
                    raise InputSyntaxError(target, 'Invalid constraint syntax')

                for vnum, vgroup in enumerate(validator.validate(coll, cspec, subject=coll_name)):
                    if len(vgroup) == 0:
                        continue
                    vsect += 1
                    vletter = letter_num(vsect)
                    sect_hdr = report.SectionHeader(title='Constraint Violations {}'.format(vletter))
                    _log.debug('Collection "{}": {:d} violations'.format(coll_name, len(vgroup)))
                    sect_hdr.add('Condition', str(vgroup.condition))
                    table = report.Table(colnames=('Id', 'TaskId', 'Field', 'Constraint', 'Value'))
                    for viol, vrec in vgroup:
                        rec_id = vrec['_id']
                        task_id = vrec['task_id']
                        if isinstance(viol.expected_value, type):
                            viol.expected_value = viol.expected_value.__name__
                        table.add((rec_id, task_id, viol.field,
                                   '{} {}'.format(viol.op, viol.expected_value),
                                   viol.got_value))
                    table.sortby('Id')
                    rpt_sect.add_section(report.ReportSection(sect_hdr, table))
            except DBError as err:
                _log.error('validating collecton {}: DB error: {}'.format(coll_name, err))
                break
            except ValueError as err:
                _log.error('validating collecton {}: {}'.format(coll_name, err))
                break
            rpt.add_section(rpt_sect)
    rpt.header.add('Elapsed time', '{:.2f}s'.format(elapsed.value))
    _log.debug('Run time: {:.2f}'.format(elapsed.value))

    # Run custom python tests
    # TODO: Eventually refactor into a real test runner.

    if args.python_module:
        # import the specified module
        python_test_module = importlib.import_module(args.python_module)
        # extract anything that begins with test_
        test_methods = [getattr(python_test_module, m) for m in dir(python_test_module) if m.startswith("test_")]

        sect_hdr = report.SectionHeader(title='Custom Validations')
        rpt_sect = report.ReportSection(sect_hdr)

        table = report.Table(colnames=('Function', 'Error'))

        for t in test_methods:
            # Run the actual test. Test should return None if successful or error string if failure
            error = t(db)
            if error:
                # Add a row to report for each test.
                table.add(('{}()'.format(t.__name__), error))

        sub_sect_hdr = report.SectionHeader(title='Python Test Functions')
        rpt_sect.add_section(report.ReportSection(sub_sect_hdr, table))

        rpt.add_section(rpt_sect)

    # Report results.
    if not args.send_on_empty and rpt.is_empty():
        _log.warn("Report is empty, skipping output")
        retcode = 1
    else:
        formatter = formatter_class()
        text = formatter.format(rpt)
        if emailer:
            if rpt.is_empty():
                emailer.subject = "{}: No Issues".format(emailer.subject)
            msgfmt = 'text/plain'
            for mt in ((report.HTMLFormatter, 'text/html'),
                       (report.JSONFormatter, 'text/plain')):
                if isinstance(formatter, mt[0]):
                    msgfmt = mt[1]
                    break
            n = emailer.send(text, fmt=msgfmt)
            if n < 1:
                _log.error("Email report not sent")
        else:
            print(text)

    return retcode

# Main
# ----


def main(cmdline=None):
    """Main program entry point.

    :param cmdline: command-line
    :return: Status, 0=ok otherwise an error
    :rtype : int
    """
    if cmdline is None:
        cmdline = sys.argv[1:]
    desc = ' '.join(__doc__.split())
    formatters = dict(json=report.JSONFormatter, html=report.HTMLFormatter,
                      md=report.MarkdownFormatter)

    # Configure parent parser for shared args.
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--verbose', '-v', dest='vb', action="count", default=0,
                               help="Print more verbose messages to standard error. Repeatable. (default=ERROR)")

    parser = argparse.ArgumentParser(description=desc, parents=[parent_parser])
    subparsers = parser.add_subparsers(description="Actions")

    # Validation command.
    subp = subparsers.add_parser("validate", help="Validate a collection", parents=[parent_parser])
    subp.set_defaults(func=command_validate, func_args=(formatters,))
    subp.add_argument('constraint', nargs='*')
    subp.add_argument('--alias', '-a', dest='aliases', action='append', type=nvpstr, metavar="ALIAS", default=[],
                      help='Alias for a field used in a constraint or condition, in form alias=name_in_db')
    subp.add_argument('--config', '-c', dest='config_file', metavar='FILE', default=None,
                      help="Configuration file for database connection. "
                           "Generate one using mgdb init --config filename.json, if necessary. "
                           "Otherwise, the code searches for a db.json. If "
                           "none is found, a no-authentication "
                           "localhost:27017/vasp database is assumed.")
    subp.add_argument('--collection', '-C', dest='coll', metavar='NAME', default=None,
                      help='Collection name for constraints given on command-line')
    subp.add_argument('--email', '-e', dest='email', default=None,
                      help='Email report to EMAIL at {from}:{to}[:host[:port[/subject]]].'
                           'The "from" and "to" are required; default host is localhost. '
                           'This information can also be in the main config file under the "_email" key, '
                           'which should be a mapping with keys "from", "to", "host", "port", and "subject".')
    subp.add_argument('--empty', dest='send_on_empty', action='store_true',
                      help="Send a report, even if it is empty (default=print warning and don't send)")
    subp.add_argument('--exonly', dest='must_exist', action='store_true', default=False,
                      help='Only show results where all fields in the constraints are also present in the record')
    subp.add_argument('--file', '-f', dest='constfile', metavar='FILE', default=None,
                      help='Main configuration file. Has constraints, and optionally email info.')
    subp.add_argument('--format', '-F', dest='report_format', metavar='FORMAT', default='html',
                      help='Report format: {} (default=html)'.format(', '.join(formatters.keys())))
    subp.add_argument('--limit', '-m', dest='limit', metavar='NUM', type=int, default=50,
                      help='Limit number of displayed constraint violations per-collection 0=no limit (50)')
    subp.add_argument('--progress', '-p', dest='progress', metavar='NUM', type=int, default=0,
                      help='Report progress every NUM invalid records found')
    subp.add_argument('--user', '-u', dest='user', metavar='NAME', default=None,
                      help='User name, for the report')
    subp.add_argument('--python-module', dest='python_module', metavar='PYTHON_MODULE', default=None,
                      help="Python module with test functions")

    # Diff command.
    subp = subparsers.add_parser("diff", help="Show difference in two collections", parents=[parent_parser])
    subp.set_defaults(func=command_diff, func_args=())
    subp.add_argument("-D", "--db", dest="rpt_db", default=None, metavar="CONFIG",
                      help="Record a JSON record of the report in the MongoDB collection configured by CONFIG, "
                           "which is a standard pymatgen-db configuration file.")
    subp.add_argument("-E", "--email", dest='rpt_email', default=None, metavar="ADDR",
                      help="Email report to ADDR, which is of the form:  "
                           "sender/receiver,[receiver2...][/subject].")
    subp.add_argument('-f', '--file', action=YamlConfig, metavar='FILE',
                      help="Read configuration from FILE instead of command line. "
                           "File is YAML (or JSON, a subset), with long option names as keys. Any "
                           "time the option takes a comma-separated list, the config file uses a real list, "
                           "and key/value pair lists are mappings. "
                           "e.g. --numeric='x=+-1.5,y=+-0.5' becomes {.., 'numeric':  ['x':'+-1.5', 'y':'+-0.5'] } "
                           "and --info=foo,bar becomes '{ .., info: ['foo', 'bar'] }'.")
    subp.add_argument("-F", "--format", dest="format", metavar="FORMAT", default=None,
                      help="Default report format: 'text', 'html', or 'json'. If not given, the format will "
                           "be determined by the output: text for console, html for email.",
                      choices=["text", "html", "json"])
    subp.add_argument("-s", "--email-server", dest="email_server", default="localhost", metavar="HOST",
                      help="Server HOST for an email report, in form hostname[:port]. Default is localhost")
    subp.add_argument("-i", "--info", help="Extra fields for records, as comma-separated list"
                                           ", e.g 'extra,fields,to_include'", dest="info",
                      default=None, type=args_list)
    subp.add_argument("-k", "--key", help="Key for matching records (default='key')", dest="key",
                      default="key")
    subp.add_argument("-m", "--missing", help="Only report keys that are in the 'old' collection, but "
                                              "not in the 'new' collection", action="store_true", dest="missonly")
    subp.add_argument("-n", "--numeric", dest="numprops", default=None, metavar="EXPR", type=args_kvp_nodup,
                      help="Fields with numeric values that must match, with a tolerance, "
                           "as a comma-separated list. "
                           "e.g., '<name1>=<expr1>, <name2>=<expr2>, ..'. "
                           "<name> is a field name, <expr> syntax is: '+-' (sign), '+-N', '+-N%%', "
                           "where N is a number. "
                           "Adding '=' (must go before '%%') compares >=, <= instead of >, <. "
                           "The '%%' means a percentage relative to the first collection's value.")
    subp.add_argument("-p", "--properties", help="Fields with properties that must match, as comma-separated list "
                                           ", e.g 'these_must,match'", dest="properties", default=None,
                      type=args_list)
    subp.add_argument("-P", "--print", help="Print report to the console", action="store_true", dest="rpt_print")
    subp.add_argument("-q", "--query", metavar="EXPR", help="Expression to filter records before diff. "
                                                            "Uses simplified constraint syntax, e.g., "
                                                            "'name = \"oscar\" and grouchiness > 3'",
                      dest="fltr")
    subp.add_argument("-u", "--url", metavar="URL", dest="rest_url",
                      help="In HTML reports, make the key into a hyperlink by prefixing with URL. "
                           "e.g., 'https://materialsproject.org/tasks/'.")
    subp.add_argument("-V", "--values", dest="changeonly", action="store_true",
                      help="Only report changes in values, not missing or added keys")
    subp.add_argument("old", help="pymatgen-db JSON config file for the 'old' collection")
    subp.add_argument("new", help="pymatgen-db JSON config file for the 'new' collection")

    args = parser.parse_args(cmdline)

    # Do some extra arg. checking for diff command, due to YAML config option.
    if hasattr(args, 'func') and args.func == command_diff:
        # Manually check types on some args
        for name, type_, opt in (("info", list, "-i/--info"), ("properties", list, "-p/--properties"),
                                 ("numprops", dict, "-n/--numeric")):
            val = getattr(args, name)
            if val is None:
                continue
            if not isinstance(val, type_):
                parser.error("Wrong type for {} argument, '{}' is not a {}"
                             .format(opt, val, type_.__name__))

    # Set logging level.
    verbosity = (logging.ERROR, logging.INFO, logging.DEBUG)[min(args.vb, 2)]
    _log.setLevel(verbosity)

    # Run subcommand.
    if hasattr(args, 'func'):
        try:
            return args.func(args, *args.func_args)
        except ArgumentError as err:
            parser.error(err.msg)
        except FunctionError as err:
            _log.critical("{}".format(err))
            return -1


if __name__ == '__main__':
    sys.exit(main())



"""
Build a derived collection with the maximum
value from each 'group' defined in the source
collection.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '5/21/14'

from matgendb.builders import core
from matgendb.builders import util
from matgendb.query_engine import QueryEngine

_log = util.get_builder_log("incr")

class MaxValueBuilder(core.Builder):
    """Example of incremental builder that requires
    some custom logic for incremental case.
    """
    def get_items(self, source=None, target=None):
        """Get all records from source collection to add to target.

        :param source: Input collection
        :type source: QueryEngine
        :param target: Output collection
        :type target: QueryEngine
        """
        self._groups = self.shared_dict()
        self._target_coll = target.collection
        self._src = source
        return source.query()

    def process_item(self, item):
        """Calculate new maximum value for each group,
        for "new" items only.
        """
        group, value = item['group'], item['value']
        if group in self._groups:
            cur_val = self._groups[group]
            self._groups[group] = max(cur_val, value)
        else:
            # New group. Could fetch old max. from target collection,
            # but for the sake of illustration recalculate it from
            # the source collection.
            self._src.tracking = False  # examine entire collection
            new_max = value
            for rec in self._src.query(criteria={'group': group},
                                       properties=['value']):
                new_max = max(new_max, rec['value'])
            self._src.tracking = True  # back to incremental mode
            # calculate new max
            self._groups[group] = new_max

    def finalize(self, errs):
        """Update target collection with calculated maximum values.
        """
        for group, value in self._groups.items():
            doc = {'group': group, 'value': value}
            self._target_coll.update({'group': group}, doc, upsert=True)
        return True


"""
Unit tests for `dbconfig` module.
"""
__author__ = 'Dan Gunter'
__date__ = '2014-04-25'

import json
import tempfile
import unittest
from matgendb.dbconfig import DBConfig, normalize_auth
from matgendb.dbconfig import ConfigurationFileError

class NormAuthTestCase(unittest.TestCase):
    """Test cases for normalize_auth().
    """
    def setUp(self):
        self.U, self.P = "joe", "secret"

    def _check(self, s):
        """Check that user/password matches expected.
        """
        self.failUnless(s["user"] == self.U)
        self.failUnless(s["password"] == self.P)

    def _check_absent(self, s):
        """Check that user/password is not present.
        """
        self.failIf("user" in s or "password" in s)

    def test_admin(self):
        s = {"admin_user": self.U, "admin_password": self.P}
        normalize_auth(s)
        self._check(s)
        s = {"admin_user": self.U, "admin_password": self.P}
        normalize_auth(s, admin=False)
        self._check_absent(s)

    def test_ro(self):
        s = {"readonly_user": self.U, "readonly_password": self.P}
        normalize_auth(s)
        self._check(s)
        s = {"readonly_user": self.U, "readonly_password": self.P}
        normalize_auth(s, readonly=False)
        self._check_absent(s)

    def test_plain(self):
        s = {"user": self.U, "password": self.P}
        normalize_auth(s)
        self._check(s)

    def test_order(self):
        s = {"readonly_user": self.U + "_1", "readonly_password": self.P,
             "admin_user": self.U + "_2", "admin_password": self.P}
        normalize_auth(s, readonly_first=False)
        self.assertEqual(s["user"], self.U + "_2")
        s = {"readonly_user": self.U + "_1", "readonly_password": self.P,
             "admin_user": self.U + "_2", "admin_password": self.P}
        normalize_auth(s, readonly_first=True)
        self.assertEqual(s["user"], self.U + "_1")


class SettingsTestCase(unittest.TestCase):
    def setUp(self):
        self.cfg = {
            "host": u"localhost",
            "port": 27017,
            "database": u"foo",
            "user": u"guy",
            "password": u"knock-knock",
            "aliases": {},
        }

    def _aliased_cfg(self):
        """Imitate authorization de-aliasing."""
        cfg = self.cfg.copy()
        cfg['readonly_password'] = cfg['password']
        cfg['readonly_user'] = cfg['user']
        return cfg

    def test_init_file(self):
        """Create DBConfig from file or path.
        """
        # sending file obj, or its path, should have same effect
        tmp = "dbconfigtest.json"
        with open(tmp, "w") as f:
            json.dump(self.cfg, f)
            f.flush()
            # reset file to beginning
        with open(tmp, "r") as f:
            d1 = DBConfig(config_file=f)
        d2 = DBConfig(config_file=tmp)
        self.assertEqual(d1.settings, d2.settings)
        self.assertEqual(d2.settings, self._aliased_cfg())

    def test_init_dict(self):
        """Create DBConfig from dict object.
        """
        d1 = DBConfig(config_dict=self.cfg)
        self.assertEqual(d1.settings, self._aliased_cfg())

    def test_init_junk_file(self):
        """Check error when creating DBConfig from bad input file.
        """
        tf = tempfile.NamedTemporaryFile("w", delete=False)
        tf.write("JUNK")
        tf.close()
        self.assertRaises(ConfigurationFileError, DBConfig, config_file=tf.name)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
This module defines a Drone to assimilate vasp data and insert it into a
Mongo database.
"""

from __future__ import division

import os
import re
import glob
import logging
import datetime
import string
import json
import socket
import numpy as np
import zlib
from fnmatch import fnmatch
from collections import OrderedDict

from pymongo import MongoClient
import gridfs

from pymatgen.apps.borg.hive import AbstractDrone
from pymatgen.analysis.local_env import VoronoiNN
from pymatgen.core.structure import Structure
from pymatgen.core.composition import Composition
from pymatgen.io.vasp import Vasprun, Incar, Kpoints, Potcar, Poscar, \
    Outcar, Oszicar
from pymatgen.io.cif import CifWriter
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.bond_valence import BVAnalyzer
from monty.io import zopen
from pymatgen.ext.matproj import MPRester
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.analysis.structure_analyzer import oxide_type
from monty.json import MontyEncoder

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "2.0.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Mar 18, 2012"


logger = logging.getLogger(__name__)


class VaspToDbTaskDrone(AbstractDrone):
    """
    VaspToDbTaskDrone assimilates directories containing vasp input to
    inserted db tasks. This drone is meant ot be used with pymatgen's
    BorgQueen to assimilate entire directory structures and insert them into
    a database using Python's multiprocessing. The current format assumes
    standard VASP relaxation runs. If you have other kinds of runs,
    you may design your own Drone class based on this one.

    There are some restrictions on the valid directory structures:

        1. There can be only one vasp run in each directory. Nested directories
           are fine.
        2. Directories designated "relax1", "relax2" are considered to be
           2 parts of an aflow style run.
        3. Directories containing vasp output with ".relax1" and ".relax2" are
           also considered as 2 parts of an aflow style run.
    """

    # Version of this db creator document.
    __version__ = "2.0.0"

    def __init__(self, host="127.0.0.1", port=27017, database="vasp",
                 user=None, password=None, collection="tasks",
                 parse_dos=False, compress_dos=False,parse_projected_eigen=False,
                 simulate_mode=False, additional_fields=None, update_duplicates=True,
                 mapi_key=None, use_full_uri=True, runs=None):
        """Constructor.

        Args:
            host:
                Hostname of database machine. Defaults to 127.0.0.1 or
                localhost.
            port:
                Port for db access. Defaults to mongo's default of 27017.
            database:
                Actual database to access. Defaults to "vasp".
            user:
                User for db access. Requires write access. Defaults to None,
                which means no authentication.
            password:
                Password for db access. Requires write access. Defaults to
                None, which means no authentication.
            collection:
                Collection to query. Defaults to "tasks".
            parse_dos:
                Whether to parse the DOS data. Options are True, False, and 'final'
                Defaults to False. If True, all dos will be inserted into a gridfs
                collection called dos_fs. If 'final', only the last calculation will
                be parsed.
            parse_projected_eigen:
                Whether to parse the element and orbital projections. Options are True,
                False, and 'final'; Defaults to False. If True, projections will be 
                parsed for each calculation. If 'final', projections for only the last
                calculation will be parsed.
            compress_dos:
                Whether to compress the DOS data. Valid options are integers 1-9,
                corresponding to zlib compression level. 1 is usually adequate.
            simulate_mode:
                Allows one to simulate db insertion without actually performing
                the insertion.
            additional_fields:
                Dict specifying additional fields to append to each doc
                inserted into the collection. For example, allows one to add
                an author or tags to a whole set of runs for example.
            update_duplicates:
                If True, if a duplicate path exists in the collection, the
                entire doc is updated. Else, duplicates are skipped.
            mapi_key:
                A Materials API key. If this key is supplied,
                the insertion code will attempt to use the Materials REST API
                to calculate stability data for inserted calculations.
                Stability assessment requires a large quantity of materials
                data. E.g., to compute the stability of a new LixFeyOz
                calculation, you need to the energies of all known
                phases in the Li-Fe-O chemical system. Using
                the Materials API, we can obtain the pre-calculated data from
                the Materials Project.

                Go to www.materialsproject.org/profile to generate or obtain
                your API key.
            use_full_uri:
                Whether to use full uri path (including hostname) for the
                directory name. Defaults to True. If False, only the abs
                path will be used.
            runs:
                Ordered list of runs to look for e.g. ["relax1", "relax2"].
                Automatically detects whether the runs are stored in the
                subfolder or file extension schema.
        """
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.collection = collection
        self.port = port
        self.simulate = simulate_mode
        if isinstance(parse_dos, str) and parse_dos != 'final':
            raise ValueError('Invalid value for parse_dos')
        if isinstance(parse_projected_eigen, str) and parse_projected_eigen != 'final':
            raise ValueError('Invalid value for parse_projected_eigen')
        self.parse_projected_eigen = parse_projected_eigen
        self.parse_dos = parse_dos
        self.compress_dos = compress_dos
        self.additional_fields = additional_fields or {}
        self.update_duplicates = update_duplicates
        self.mapi_key = mapi_key
        self.use_full_uri = use_full_uri
        self.runs = runs or ["relax1", "relax2"]
        if not simulate_mode:
            conn = MongoClient(self.host, self.port)
            db = conn[self.database]
            if self.user:
                db.authenticate(self.user, self.password)
            if db.counter.find({"_id": "taskid"}).count() == 0:
                db.counter.insert_one({"_id": "taskid", "c": 1})

    def assimilate(self, path):
        """
        Parses vasp runs. Then insert the result into the db. and return the
        task_id or doc of the insertion.

        Returns:
            If in simulate_mode, the entire doc is returned for debugging
            purposes. Else, only the task_id of the inserted doc is returned.
        """
        try:
            d = self.get_task_doc(path)
            if self.mapi_key is not None and d["state"] == "successful":
                self.calculate_stability(d)
            tid = self._insert_doc(d)
            return tid
        except Exception as ex:
            import traceback
            logger.error(traceback.format_exc())
            return False

    def calculate_stability(self, d):
        m = MPRester(self.mapi_key)
        functional = d["pseudo_potential"]["functional"]
        syms = ["{} {}".format(functional, l)
                for l in d["pseudo_potential"]["labels"]]
        entry = ComputedEntry(Composition(d["unit_cell_formula"]),
                              d["output"]["final_energy"],
                              parameters={"hubbards": d["hubbards"],
                                          "potcar_symbols": syms})
        data = m.get_stability([entry])[0]
        for k in ("e_above_hull", "decomposes_to"):
            d["analysis"][k] = data[k]

    def get_task_doc(self, path):
        """
        Get the entire task doc for a path, including any post-processing.
        """
        logger.info("Getting task doc for base dir :{}".format(path))
        files = os.listdir(path)
        vasprun_files = OrderedDict()
        if "STOPCAR" in files:
            #Stopped runs. Try to parse as much as possible.
            logger.info(path + " contains stopped run")
        for r in self.runs:
            if r in files: #try subfolder schema
                for f in os.listdir(os.path.join(path, r)):
                    if fnmatch(f, "vasprun.xml*"):
                        vasprun_files[r] = os.path.join(r, f)
            else: #try extension schema
                for f in files:
                    if fnmatch(f, "vasprun.xml.{}*".format(r)):
                        vasprun_files[r] = f
        if len(vasprun_files) == 0:
            for f in files: #get any vasprun from the folder
                if fnmatch(f, "vasprun.xml*") and \
                        f not in vasprun_files.values():
                    vasprun_files['standard'] = f

        if len(vasprun_files) > 0:
            d = self.generate_doc(path, vasprun_files)
            if not d:
                d = self.process_killed_run(path)
            self.post_process(path, d)
        elif (not (path.endswith("relax1") or
              path.endswith("relax2"))) and contains_vasp_input(path):
            #If not Materials Project style, process as a killed run.
            logger.warning(path + " contains killed run")
            d = self.process_killed_run(path)
            self.post_process(path, d)
        else:
            raise ValueError("No VASP files found!")

        return d

    def _insert_doc(self, d):
        if not self.simulate:
            # Perform actual insertion into db. Because db connections cannot
            # be pickled, every insertion needs to create a new connection
            # to the db.
            conn = MongoClient(self.host, self.port)
            db = conn[self.database]
            if self.user:
                db.authenticate(self.user, self.password)
            coll = db[self.collection]

            # Insert dos data into gridfs and then remove it from the dict.
            # DOS data tends to be above the 4Mb limit for mongo docs. A ref
            # to the dos file is in the dos_fs_id.
            result = coll.find_one({"dir_name": d["dir_name"]},
                                   ["dir_name", "task_id"])
            if result is None or self.update_duplicates:
                if self.parse_dos and "calculations" in d:
                    for calc in d["calculations"]:
                        if "dos" in calc:
                            dos = json.dumps(calc["dos"], cls=MontyEncoder)
                            if self.compress_dos:
                                dos = zlib.compress(dos.encode('utf-8'),
                                                    self.compress_dos)
                                calc["dos_compression"] = "zlib"
                            fs = gridfs.GridFS(db, "dos_fs")
                            dosid = fs.put(dos)
                            calc["dos_fs_id"] = dosid
                            del calc["dos"]

                d["last_updated"] = datetime.datetime.today()
                if result is None:
                    if ("task_id" not in d) or (not d["task_id"]):
                        result = db.counter.find_one_and_update(
                            filter={"_id": "taskid"},
                            update={"$inc": {"c": 1}}
                            )
                        d["task_id"] = result["c"]
                    logger.info("Inserting {} with taskid = {}"
                                .format(d["dir_name"], d["task_id"]))
                elif self.update_duplicates:
                    d["task_id"] = result["task_id"]
                    logger.info("Updating {} with taskid = {}"
                                .format(d["dir_name"], d["task_id"]))

                coll.update_one({"dir_name": d["dir_name"]}, {"$set": d},
                                upsert=True)
                return d["task_id"]
            else:
                logger.info("Skipping duplicate {}".format(d["dir_name"]))
        else:
            d["task_id"] = 0
            logger.info("Simulated insert into database for {} with task_id {}"
                        .format(d["dir_name"], d["task_id"]))
            return d

    def post_process(self, dir_name, d):
        """
        Simple post-processing for various files other than the vasprun.xml.
        Called by generate_task_doc. Modify this if your runs have other
        kinds of processing requirements.

        Args:
            dir_name:
                The dir_name.
            d:
                Current doc generated.
        """
        logger.info("Post-processing dir:{}".format(dir_name))

        fullpath = os.path.abspath(dir_name)

        # VASP input generated by pymatgen's alchemy has a
        # transformations.json file that keeps track of the origin of a
        # particular structure. This is extremely useful for tracing back a
        # result. If such a file is found, it is inserted into the task doc
        # as d["transformations"]
        transformations = {}
        filenames = glob.glob(os.path.join(fullpath, "transformations.json*"))
        if len(filenames) >= 1:
            with zopen(filenames[0], "rt") as f:
                transformations = json.load(f)
                try:
                    m = re.match("(\d+)-ICSD",
                                 transformations["history"][0]["source"])
                    if m:
                        d["icsd_id"] = int(m.group(1))
                except Exception as ex:
                    logger.warning("Cannot parse ICSD from transformations "
                                   "file.")
                    pass
        else:
            logger.warning("Transformations file does not exist.")

        other_parameters = transformations.get("other_parameters")
        new_tags = None
        if other_parameters:
            # We don't want to leave tags or authors in the
            # transformations file because they'd be copied into
            # every structure generated after this one.
            new_tags = other_parameters.pop("tags", None)
            new_author = other_parameters.pop("author", None)
            if new_author:
                d["author"] = new_author
            if not other_parameters:  # if dict is now empty remove it
                transformations.pop("other_parameters")

        d["transformations"] = transformations

        # Calculations done using custodian has a custodian.json,
        # which tracks the jobs performed and any errors detected and fixed.
        # This is useful for tracking what has actually be done to get a
        # result. If such a file is found, it is inserted into the task doc
        # as d["custodian"]
        filenames = glob.glob(os.path.join(fullpath, "custodian.json*"))
        if len(filenames) >= 1:
            with zopen(filenames[0], "rt") as f:
                d["custodian"] = json.load(f)

        # Parse OUTCAR for additional information and run stats that are
        # generally not in vasprun.xml.
        try:
            run_stats = {}
            for filename in glob.glob(os.path.join(fullpath, "OUTCAR*")):
                outcar = Outcar(filename)
                i = 1 if re.search("relax2", filename) else 0
                taskname = "relax2" if re.search("relax2", filename) else \
                    "relax1"
                d["calculations"][i]["output"]["outcar"] = outcar.as_dict()
                run_stats[taskname] = outcar.run_stats
        except:
            logger.error("Bad OUTCAR for {}.".format(fullpath))

        try:
            overall_run_stats = {}
            for key in ["Total CPU time used (sec)", "User time (sec)",
                        "System time (sec)", "Elapsed time (sec)"]:
                overall_run_stats[key] = sum([v[key]
                                              for v in run_stats.values()])
            run_stats["overall"] = overall_run_stats
        except:
            logger.error("Bad run stats for {}.".format(fullpath))

        d["run_stats"] = run_stats

        #Convert to full uri path.
        if self.use_full_uri:
            d["dir_name"] = get_uri(dir_name)

        if new_tags:
            d["tags"] = new_tags

        logger.info("Post-processed " + fullpath)

    def process_killed_run(self, dir_name):
        """
        Process a killed vasp run.
        """
        fullpath = os.path.abspath(dir_name)
        logger.info("Processing Killed run " + fullpath)
        d = {"dir_name": fullpath, "state": "killed", "oszicar": {}}

        for f in os.listdir(dir_name):
            filename = os.path.join(dir_name, f)
            if fnmatch(f, "INCAR*"):
                try:
                    incar = Incar.from_file(filename)
                    d["incar"] = incar.as_dict()
                    d["is_hubbard"] = incar.get("LDAU", False)
                    if d["is_hubbard"]:
                        us = np.array(incar.get("LDAUU", []))
                        js = np.array(incar.get("LDAUJ", []))
                        if sum(us - js) == 0:
                            d["is_hubbard"] = False
                            d["hubbards"] = {}
                    else:
                        d["hubbards"] = {}
                    if d["is_hubbard"]:
                        d["run_type"] = "GGA+U"
                    elif incar.get("LHFCALC", False):
                        d["run_type"] = "HF"
                    else:
                        d["run_type"] = "GGA"
                except Exception as ex:
                    print(str(ex))
                    logger.error("Unable to parse INCAR for killed run {}."
                                 .format(dir_name))
            elif fnmatch(f, "KPOINTS*"):
                try:
                    kpoints = Kpoints.from_file(filename)
                    d["kpoints"] = kpoints.as_dict()
                except:
                    logger.error("Unable to parse KPOINTS for killed run {}."
                                 .format(dir_name))
            elif fnmatch(f, "POSCAR*"):
                try:
                    s = Poscar.from_file(filename).structure
                    comp = s.composition
                    el_amt = s.composition.get_el_amt_dict()
                    d.update({"unit_cell_formula": comp.as_dict(),
                              "reduced_cell_formula": comp.to_reduced_dict,
                              "elements": list(el_amt.keys()),
                              "nelements": len(el_amt),
                              "pretty_formula": comp.reduced_formula,
                              "anonymous_formula": comp.anonymized_formula,
                              "nsites": comp.num_atoms,
                              "chemsys": "-".join(sorted(el_amt.keys()))})
                    d["poscar"] = s.as_dict()
                except:
                    logger.error("Unable to parse POSCAR for killed run {}."
                                 .format(dir_name))
            elif fnmatch(f, "POTCAR*"):
                try:
                    potcar = Potcar.from_file(filename)
                    d["pseudo_potential"] = {
                        "functional": potcar.functional.lower(),
                        "pot_type": "paw",
                        "labels": potcar.symbols}
                except:
                    logger.error("Unable to parse POTCAR for killed run in {}."
                                 .format(dir_name))
            elif fnmatch(f, "OSZICAR"):
                try:
                    d["oszicar"]["root"] = \
                        Oszicar(os.path.join(dir_name, f)).as_dict()
                except:
                    logger.error("Unable to parse OSZICAR for killed run in {}."
                                 .format(dir_name))
            elif re.match("relax\d", f):
                if os.path.exists(os.path.join(dir_name, f, "OSZICAR")):
                    try:
                        d["oszicar"][f] = Oszicar(
                            os.path.join(dir_name, f, "OSZICAR")).as_dict()
                    except:
                        logger.error("Unable to parse OSZICAR for killed "
                                     "run in {}.".format(dir_name))
        return d

    def process_vasprun(self, dir_name, taskname, filename):
        """
        Process a vasprun.xml file.
        """
        vasprun_file = os.path.join(dir_name, filename)
        if self.parse_projected_eigen and (self.parse_projected_eigen != 'final' or \
                             taskname == self.runs[-1]):
            parse_projected_eigen = True
        else:
            parse_projected_eigen = False
        r = Vasprun(vasprun_file,parse_projected_eigen=parse_projected_eigen)
        d = r.as_dict()
        d["dir_name"] = os.path.abspath(dir_name)
        d["completed_at"] = \
            str(datetime.datetime.fromtimestamp(os.path.getmtime(
                vasprun_file)))
        d["cif"] = str(CifWriter(r.final_structure))
        d["density"] = r.final_structure.density
        if self.parse_dos and (self.parse_dos != 'final' \
                               or taskname == self.runs[-1]):
            try:
                d["dos"] = r.complete_dos.as_dict()
            except Exception:
                logger.warning("No valid dos data exist in {}.\n Skipping dos"
                               .format(dir_name))
        if taskname == "relax1" or taskname == "relax2":
            d["task"] = {"type": "aflow", "name": taskname}
        else:
            d["task"] = {"type": taskname, "name": taskname}
        d["oxide_type"] = oxide_type(r.final_structure)
        return d

    def generate_doc(self, dir_name, vasprun_files):
        """
        Process aflow style runs, where each run is actually a combination of
        two vasp runs.
        """
        try:
            fullpath = os.path.abspath(dir_name)
            # Defensively copy the additional fields first.  This is a MUST.
            # Otherwise, parallel updates will see the same object and inserts
            # will be overridden!!
            d = {k: v for k, v in self.additional_fields.items()}
            d["dir_name"] = fullpath
            d["schema_version"] = VaspToDbTaskDrone.__version__
            d["calculations"] = [
                self.process_vasprun(dir_name, taskname, filename)
                for taskname, filename in vasprun_files.items()]
            d1 = d["calculations"][0]
            d2 = d["calculations"][-1]

            # Now map some useful info to the root level.
            for root_key in ["completed_at", "nsites", "unit_cell_formula",
                             "reduced_cell_formula", "pretty_formula",
                             "elements", "nelements", "cif", "density",
                             "is_hubbard", "hubbards", "run_type"]:
                d[root_key] = d2[root_key]
            d["chemsys"] = "-".join(sorted(d2["elements"]))

            # store any overrides to the exchange correlation functional
            xc = d2["input"]["incar"].get("GGA")
            if xc:
                xc = xc.upper()
            d["input"] = {"crystal": d1["input"]["crystal"],
                          "is_lasph": d2["input"]["incar"].get("LASPH", False),
                          "potcar_spec": d1["input"].get("potcar_spec"),
                          "xc_override": xc}
            vals = sorted(d2["reduced_cell_formula"].values())
            d["anonymous_formula"] = {string.ascii_uppercase[i]: float(vals[i])
                                      for i in range(len(vals))}
            d["output"] = {
                "crystal": d2["output"]["crystal"],
                "final_energy": d2["output"]["final_energy"],
                "final_energy_per_atom": d2["output"]["final_energy_per_atom"]}
            d["name"] = "aflow"
            p = d2["input"]["potcar_type"][0].split("_")
            pot_type = p[0]
            functional = "lda" if len(pot_type) == 1 else "_".join(p[1:])
            d["pseudo_potential"] = {"functional": functional.lower(),
                                     "pot_type": pot_type.lower(),
                                     "labels": d2["input"]["potcar"]}
            if len(d["calculations"]) == len(self.runs) or \
                    list(vasprun_files.keys())[0] != "relax1":
                d["state"] = "successful" if d2["has_vasp_completed"] \
                    else "unsuccessful"
            else:
                d["state"] = "stopped"
            d["analysis"] = get_basic_analysis_and_error_checks(d)

            sg = SpacegroupAnalyzer(Structure.from_dict(d["output"]["crystal"]),
                                    0.1)
            d["spacegroup"] = {"symbol": sg.get_space_group_symbol(),
                               "number": sg.get_space_group_number(),
                               "point_group": sg.get_point_group_symbol(),
                               "source": "spglib",
                               "crystal_system": sg.get_crystal_system(),
                               "hall": sg.get_hall()}
            d["oxide_type"] = d2["oxide_type"]
            d["last_updated"] = datetime.datetime.today()
            return d
        except Exception as ex:
            import traceback
            print(traceback.format_exc())
            logger.error("Error in " + os.path.abspath(dir_name) +
                         ".\n" + traceback.format_exc())

            return None

    def get_valid_paths(self, path):
        """
        There are some restrictions on the valid directory structures:

        1. There can be only one vasp run in each directory. Nested directories
           are fine.
        2. Directories designated "relax1", "relax2" are considered to be 2
           parts of an aflow style run.
        3. Directories containing vasp output with ".relax1" and ".relax2" are
           also considered as 2 parts of an aflow style run.
        """
        (parent, subdirs, files) = path
        if set(self.runs).intersection(subdirs):
            return [parent]
        if not any([parent.endswith(os.sep + r) for r in self.runs]) and \
                len(glob.glob(os.path.join(parent, "vasprun.xml*"))) > 0:
            return [parent]
        return []

    def convert(self, d):
        return d

    def __str__(self):
        return "VaspToDbDictDrone"

    @classmethod
    def from_dict(cls, d):
        return cls(**d["init_args"])

    def as_dict(self):
        init_args = {"host": self.host, "port": self.port,
                     "database": self.database, "user": self.user,
                     "password": self.password,
                     "collection": self.collection,
                     "parse_dos": self.parse_dos,
                     "simulate_mode": self.simulate,
                     "additional_fields": self.additional_fields,
                     "update_duplicates": self.update_duplicates}
        output = {"name": self.__class__.__name__,
                  "init_args": init_args, "version": __version__}
        return output


def get_basic_analysis_and_error_checks(d, max_force_threshold=0.5,
                                        volume_change_threshold=0.2):

    initial_vol = d["input"]["crystal"]["lattice"]["volume"]
    final_vol = d["output"]["crystal"]["lattice"]["volume"]
    delta_vol = final_vol - initial_vol
    percent_delta_vol = delta_vol / initial_vol
    coord_num = get_coordination_numbers(d)
    calc = d["calculations"][-1]
    gap = calc["output"]["bandgap"]
    cbm = calc["output"]["cbm"]
    vbm = calc["output"]["vbm"]
    is_direct = calc["output"]["is_gap_direct"]

    warning_msgs = []
    error_msgs = []

    if abs(percent_delta_vol) > volume_change_threshold:
        warning_msgs.append("Volume change > {}%"
                            .format(volume_change_threshold * 100))

    bv_struct = Structure.from_dict(d["output"]["crystal"])
    try:
        bva = BVAnalyzer()
        bv_struct = bva.get_oxi_state_decorated_structure(bv_struct)
    except ValueError as e:
        logger.error("Valence cannot be determined due to {e}."
                     .format(e=e))
    except Exception as ex:
        logger.error("BVAnalyzer error {e}.".format(e=str(ex)))

    max_force = None
    if d["state"] == "successful" and \
            d["calculations"][0]["input"]["parameters"].get("NSW", 0) > 0:
        # handle the max force and max force error
        max_force = max([np.linalg.norm(a)
                        for a in d["calculations"][-1]["output"]
                        ["ionic_steps"][-1]["forces"]])

        if max_force > max_force_threshold:
            error_msgs.append("Final max force exceeds {} eV"
                              .format(max_force_threshold))
            d["state"] = "error"

        s = Structure.from_dict(d["output"]["crystal"])
        if not s.is_valid():
            error_msgs.append("Bad structure (atoms are too close!)")
            d["state"] = "error"

    return {"delta_volume": delta_vol,
            "max_force": max_force,
            "percent_delta_volume": percent_delta_vol,
            "warnings": warning_msgs,
            "errors": error_msgs,
            "coordination_numbers": coord_num,
            "bandgap": gap, "cbm": cbm, "vbm": vbm,
            "is_gap_direct": is_direct,
            "bv_structure": bv_struct.as_dict()}


def contains_vasp_input(dir_name):
    """
    Checks if a directory contains valid VASP input.

    Args:
        dir_name:
            Directory name to check.

    Returns:
        True if directory contains all four VASP input files (INCAR, POSCAR,
        KPOINTS and POTCAR).
    """
    for f in ["INCAR", "POSCAR", "POTCAR", "KPOINTS"]:
        if not os.path.exists(os.path.join(dir_name, f)) and \
                not os.path.exists(os.path.join(dir_name, f + ".orig")):
            return False
    return True


def get_coordination_numbers(d):
    """
    Helper method to get the coordination number of all sites in the final
    structure from a run.

    Args:
        d:
            Run dict generated by VaspToDbTaskDrone.

    Returns:
        Coordination numbers as a list of dict of [{"site": site_dict,
        "coordination": number}, ...].
    """
    structure = Structure.from_dict(d["output"]["crystal"])
    f = VoronoiNN()
    cn = []
    for i, s in enumerate(structure.sites):
        try:
            n = f.get_cn(structure, i)
            number = int(round(n))
            cn.append({"site": s.as_dict(), "coordination": number})
        except Exception:
            logger.error("Unable to parse coordination errors")
    return cn


def get_uri(dir_name):
    """
    Returns the URI path for a directory. This allows files hosted on
    different file servers to have distinct locations.

    Args:
        dir_name:
            A directory name.

    Returns:
        Full URI path, e.g., fileserver.host.com:/full/path/of/dir_name.
    """
    fullpath = os.path.abspath(dir_name)
    try:
        hostname = socket.gethostbyaddr(socket.gethostname())[0]
    except:
        hostname = socket.gethostname()
    return "{}:{}".format(hostname, fullpath)



"""
Unit tests for `dbgroup` module.
"""
__author__ = "Dan Gunter"
__copyright__ = "Copyright 2014, The Materials Project"
__email__ = "dkgunter@lbl.gov"
__date__ = "2014-04-29"

import json
import mongomock
import os
import tempfile
import unittest
from matgendb.dbgroup import ConfigGroup
from matgendb import dbconfig

_opj = os.path.join

mockdb = mongomock.MongoClient()
doc = {"hello": "world"}
mockdb.testdb.data.insert(doc)
# add some nested collections
mockcoll = [
    'data.a1',
    'data.a1.b1',
    'data.a1.b2',
    'data.a2'
]
[mockdb.testdb[c].insert(doc) for c in mockcoll]

class MockQueryEngine:
    def __init__(self, **kwargs):
        self.kw = kwargs

    def __eq__(self, other):
        return other.kw == self.kw

    @property
    def db(self):
        return mockdb.testdb

    @property
    def collection(self):
        return mockdb.testdb[self.kw['collection']]

    @property
    def db(self):
        class DB:
            _c = mockcoll
            def collection_names(self, x=None):
                return self._c
        return DB()

class Cfg:
    def __init__(self, v):
        self.settings = {'collection': v}
        self.collection = v
    def copy(self):
        return Cfg(self.collection)

class ConfigGroupTestCase(unittest.TestCase):
    def setUp(self):
        self.g = ConfigGroup(qe_class=MockQueryEngine)
        self.configs = [Cfg("qe{:d}".format(i)) for i in range(5)]

    def test_add(self):
        """ConfigGroup add and lookup
        """
        keys = ["foo", "bar", "foo.a", "foo.b"]
        expect = {}
        for i, k in enumerate(keys):
            self.g.add(k, self.configs[i])
            expect[k] = MockQueryEngine(**self.configs[i].settings)
        self.assertEqual(self.g["foo"], expect["foo"])
        self.assertEqual(self.g["bar"], expect["bar"])
        self.assertEqual(self.g["bar*"], {"bar": expect["bar"]})
        self.assertEqual(self.g["foo.a"], expect["foo.a"])
        self.assertEqual(self.g["foo.*"],
                         {"foo.a": expect["foo.a"], "foo.b": expect["foo.b"]})

    def test_add_path(self):
        """Add set of query engines from a path.
        """
        # directory of pretend configs
        d = tempfile.mkdtemp()
        try:
            # fill with some configs
            c = {}
            for root in ("foo", "bar"):
                for sub in ("a", "b.1", "b.2"):
                    config = {
                        dbconfig.DB_KEY: root,
                        dbconfig.COLL_KEY: sub
                    }
                    filename = "mg_core_{}_{}.json".format(root, sub)
                    with open(_opj(d, filename), "w") as fp:
                        json.dump(config, fp)
                    c["{}.{}".format(root, sub)] = config
            # read them
            self.g.add_path(d)
            # check all were added
            self.assertEqual(sorted(self.g.keys()), sorted(c.keys()))
            # check one
            qe1 = self.g["foo.a"].kw
            c1 = c["foo.a"]
            self.assertTrue(dict_subset(c1, qe1))
            # check with prefix
            self.g.set_prefix("foo.b")
            self.assertTrue(dict_subset(c["foo.b.1"], self.g["1"].kw))
            self.assertRaises(KeyError, self.g.__getitem__, "bla")
            # check list with prefix
            gkeys = sorted(self.g["*"].keys())
            self.assertEqual(gkeys, ["foo.b.1", "foo.b.2"])
            # check list w/o prefix
            self.g.set_prefix()
            gkeys = sorted(self.g["bar.b.*"].keys())
            self.assertEqual(gkeys, ["bar.b.1", "bar.b.2"])
        finally:
            # rm -r $d
            for f in os.listdir(d):
                os.unlink(os.path.join(d, f))
            os.rmdir(d)

    def test_uncache(self):
        """Remove cached query engine(s) from ConfigGroup.
        """
        keys = ("foo.a", "foo", "bar")
        for i in range(len(keys)):
            self.g.add(keys[i], self.configs[i])
        # force instantiation/caching
        for i in range(len(keys)):
            self.g[keys[i]]
        left_behind = self.g[keys[2]]
        # remove all foo from cache
        self.g.uncache("foo*")
        # check that they are not cached
        for i in range(2):
            self.assertRaises(KeyError, self.g._cached.__getitem__, keys[i])
        # check that un-removed remain
        self.assertEqual(self.g[keys[2]], left_behind)

    def test_expand(self):
        """Add multiple collections at once with 'expand'.
        """
        self.g.add("foo", Cfg("data"), expand=True)
        # check that data.* got added as foo.*
        keys = set(self.g.keys())
        expect = set(["foo"] + [f.replace("data", "foo") for f in mockcoll])
        self.assertEqual(expect, keys)

def dict_subset(a, b):
    for k in a.keys():
        if k not in b or b[k] != a[k]:
            return False
    return True

if __name__ == '__main__':
    unittest.main()



"""
Incremental builders.

## High-level API ##

The main class is TrackedQueryEngine. Usage example::

    from matgendb.builders.incr import *
    qe = TrackedQueryEngine(track_operation=Operation.copy,
                            track_field='_id', ...other kw...)
    # That's it! Use as you normally would.
    # The collection 'inside' the qe will be tracked, so
    # when you are done using the QE instance, you can call either
    qe.set_mark()
    # or
    qe.collection.set_mark()
    # Just be sure not to set the `qe.collection` attribute directly,
    # always use `qe.collection_name = "new_name" (which calls a setter)`.

## Low-level API ##

The main classes are Mark and CollectionTracker. Usage example::

    from matgendb.builders.incr import *
    collection = pymongo.MongoClient().mydb.mycollection
    # Init tracker
    tracker = CollectionTracker(collection)
    # Save mark for copy operation
    tracker.save(Mark(collection, Operation.copy, field='_id'))
    # Retrieve mark for copy operation
    mark = tracker.retrieve(Operation.copy, field='_id')
    # Update mark position, and print mark
    print(mark.update().as_dict())

"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '4/11/14'

from abc import abstractmethod, ABCMeta
import pymongo
from enum import Enum
from matgendb.query_engine import QueryEngine
from matgendb.builders import util as bld_util

# Logging

_log = bld_util.get_builder_log("incr")

# Exceptions

class DBError(Exception):
    """Generic database error.
    """
    pass


class NoTrackingCollection(Exception):
    """Raised if no tracking collection is present,
    but some operation is requested on that collection.
    """
    pass


## --------------
## High level API
## --------------

class TrackingInterface(metaclass=ABCMeta):
    @abstractmethod
    def set_mark(self):
        """Set the mark to the current end of the collection. This is saved in the database
        so it is available for later operations.
        """
        pass

class UnTrackedQueryEngine(QueryEngine, TrackingInterface):
    """A QE that has the interface for tracking, but does nothing for it.
    Allows for callers to do same operations regardless of whether tracking is
    activated or not.
    """
    def set_mark(self):
        """Does nothing and returns None.
        """
        return


class TrackedQueryEngine(QueryEngine, TrackingInterface):
    """A QueryEngine subclass that only examines records
    past the last 'mark' that was set for the
    given operation and field.

    The concrete result is that for an object, ``t``, the
    ``t.find()`` method will start *past* the "mark".

    To go around this transparent change, use the ``t.findall``
    instead.
    """
    def __init__(self, track_operation=None, track_field=None, **kwargs):
        """Constructor.
        """
        self._tracking_off = False
        # Set these first because QueryEngine.__init__ calls overridden `collection_name setter()`.
        assert track_field
        self._t_op, self._t_field = track_operation, track_field
        self.collection = None
        # Now init parent
        QueryEngine.__init__(self, **kwargs)

    @property
    def tracking(self):
        """Whether tracking is really enabled."""
        return not self._tracking_off

    @tracking.setter
    def tracking(self, is_tracked):
        self._tracking_off = not is_tracked
        if self.collection is not None:
            self.collection.set_tracking(is_tracked)

    @property
    def collection_name(self):
        """Override base class to make this a tracked collection.
        See :meth:`@collection_name.setter`
        """
        return self._collection_name

    @collection_name.setter
    def collection_name(self, value):
        """Switch to another collection.
        Note that you may have to set the aliases and default properties if the
        schema of the new collection differs from the current collection.
        """
        self._collection_name = value
        self._mongo_coll = self.db[value]
        self.collection = TrackedCollection(self._mongo_coll, operation=self._t_op,
                                            field=self._t_field)

    def set_mark(self):
        """See :meth:`TrackingInterface.set_mark`
        """
        assert self.collection
        self.collection.set_mark()

class TrackedCollection:
    """Wrapper on a pymongo collection to make `find' operations start
    after the "tracking" mark.
    """
    def __init__(self, coll, operation=None, field=None):
        self._coll, self._coll_find = coll, coll.find
        self._tracking_off = False
        self._tracker = CollectionTracker(coll, create=True)
        self._mark = self._tracker.retrieve(operation=operation, field=field)

    def set_tracking(self, is_tracked):
        """Set whether find() transparently starts from after the
        mark (True), or whether it ignores the tracking info and acts like
        a normal ``find()`` operation (False).

        :param is_tracked: Whether to use tracking info
        :return:
        """
        self._tracking_off = not is_tracked

    def findall(self, *args, **kwargs):
        """Call non-tracked ``find()`` operation with same args.
        """
        return self._coll.find(*args, **kwargs)

    def __getattr__(self, item):
        if item == 'find':
            # monkey-patch the find() method in the collection object
            return self.tracked_find
        else:
            return getattr(self._coll, item)

    def __str__(self):
        return "Tracked collection ({})".format(self._coll)

    def tracked_find(self, *args, **kwargs):
        """Replacement for regular ``find()``.
        """
        _log.info("tracked_find.begin")
        # if tracking is off, just call find (ie do nothing)
        if self._tracking_off:
            _log.info("tracked_find.end, tracking=off")
            return self._coll_find(*args, **kwargs)
        # otherwise do somethin' real
        # fish 'filter' out of args or kwargs
        if len(args) > 0:
            filt = args[0]
        else:
            if 'filter' not in kwargs:
                kwargs['filter'] = {}
            filt = kwargs['filter']
        # update filter with tracker query
        filt.update(self._mark.query)
        # delegate to "real" find()
        _log.info("tracked_find.end, call: {}.find(args={} kwargs={})".format(self._coll.name, args, kwargs))
        return self._coll_find(*args, **kwargs)

    def set_mark(self):
        self._tracker.save(self._mark.update())

# TODO: TrackedFileset -- Same basic idea with one or more files in a directory.
# TODO: This would enable the incremental interface to work just as well with loading
# TODO: from files into the DB as it does with DB -> DB operations.


## --------------
## Low level API
## --------------

class Operation(Enum):
    """Enumeration of collection operations.
    """
    copy = 1
    build = 2
    other = 99


class Mark:
    """The position in a collection for the last record that was
    processed by a given operation.
    """
    # Fields for JSON representation
    FLD_OP, FLD_MARK, FLD_FLD = "operation", "mark", "field"

    def __init__(self, collection=None, operation=None, field=None, pos=None):
        """Constructor.

        :param collection: Collection to track
        :type collection: pymongo.collection.Collection
        :param operation: Operation for this mark.
        :type operation: Operation
        :param field: Name of field to determine highest record
        :type field: str
        :param pos: Optional query for position
        :type pos: dict
        """
        self._c = collection
        self._op = operation
        self._fld = field
        assert self._fld
        if pos:
            self._pos = pos
        else:
            self._pos = self._empty_pos()

    def update(self):
        """Update the position of the mark in the collection.

        :return: this object, for chaining
        :rtype: Mark
        """
        rec = self._c.find_one({}, {self._fld: 1}, sort=[(self._fld, -1)], limit=1)
        if rec is None:
            self._pos = self._empty_pos()
        elif not self._fld in rec:
            _log.error("Tracking field not found. field={} collection={}"
                       .format(self._fld, self._c.name))
            _log.warn("Continuing without tracking")
            self._pos = self._empty_pos()
        else:
            self._pos = {self._fld: rec[self._fld]}
        return self

    def _empty_pos(self):
        return {self._fld: None}

    @property
    def pos(self):
        return self._pos

    def as_dict(self):
        """Representation as a dict for JSON serialization.
        """
        return {self.FLD_OP: self._op.name,
                self.FLD_MARK: self._pos,
                self.FLD_FLD: self._fld}

    to_dict = as_dict   # synonym

    @classmethod
    def from_dict(cls, coll, d):
        """Construct from dict

        :param coll: Collection for the mark
        :param d: Input
        :type d: dict
        :return: new instance
        :rtype: Mark
        """
        return Mark(collection=coll, operation=Operation[d[cls.FLD_OP]],
                    pos=d[cls.FLD_MARK], field=d[cls.FLD_FLD])

    @property
    def query(self):
        """A mongdb query expression to find all records with higher values
        for this mark's fields in the collection.

        :rtype: dict
        """
        q = {}
        for field, value in self._pos.items():
            if value is None:
                q.update({field: {'$exists': True}})
            else:
                q.update({field: {'$gt': value}})
        return q


class CollectionTracker:
    """Track which records are 'new' in a MongoDB collection
    with respect to a given operation.
    """
    #: Sub-collection name, added as ".<TRACKING_NAME>" to the
    #: name of the target collection to create the tracking collection.
    TRACKING_NAME = 'tracker'

    def __init__(self, coll, create=True):
        """Constructor.

        :param coll: Collection to track
        :type coll: pymongo.collection.Collection
        :param create: Create tracking collection, if not present. Otherwise the tracking
                       collection can be manually created with :meth:`create`, later.
                       If the collection is not created, :meth:`save` and :meth:`retrieve` will
                       raise a `NoTrackingCollection` exception.
        :type create: bool
        """
        self.collection, self.db = coll, coll.database
        trk = self.tracking_collection_name
        if not create and trk not in self.db.collection_names(False):
            self._track = None
        else:
            self._track = self.db[trk]

    @property
    def tracking_collection_name(self):
        return self.collection.name + '.' + self.TRACKING_NAME

    @property
    def tracking_collection(self):
        """Return current tracking collection, or None if it does not exist.
        """
        return self._track

    def create(self):
        """Create tracking collection.
        Does nothing if tracking collection already exists.
        """
        if self._track is None:
            self._track = self.db[self.tracking_collection_name]

    def save(self, mark):
        """Save a position in this collection.

        :param mark: The position to save
        :type mark: Mark
        :raises: DBError, NoTrackingCollection
        """
        self._check_exists()
        obj = mark.as_dict()
        try:
            # Make a 'filter' to find/update existing record, which uses
            # the field name and operation (but not the position).
            filt = {k: obj[k] for k in (mark.FLD_FLD, mark.FLD_OP)}
            _log.debug("save: upsert-spec={} upsert-obj={}".format(filt, obj))
            self._track.update(filt, obj, upsert=True)
        except pymongo.errors.PyMongoError as err:
            raise DBError("{}".format(err))

    def retrieve(self, operation, field=None):
        """Retrieve a position in this collection.

        :param operation: Name of an operation
        :type operation: :class:`Operation`
        :param field: Name of field for sort order
        :type field: str
        :return: The position for this operation
        :rtype: Mark
        :raises: NoTrackingCollection
        """
        obj = self._get(operation, field)
        if obj is None:
            # empty Mark instance
            return Mark(collection=self.collection, operation=operation, field=field)
        return Mark.from_dict(self.collection, obj)

    def _get(self, operation, field):
        """Get tracked position for a given operation and field."""
        self._check_exists()
        query = {Mark.FLD_OP: operation.name,
                 Mark.FLD_MARK + "." + field: {"$exists": True}}
        return self._track.find_one(query)

    def _check_exists(self):
        """Check whether the tracked collection exists at all.
        If not, raises NoTrackingCollection
        """
        if self._track is None:
            raise NoTrackingCollection()



"""
Description.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '2/21/13'


from email.mime.text import MIMEText
import json
from operator import itemgetter
import smtplib
#
from .util import DoesLogging, JsonWalker
from ..util import MongoJSONEncoder
from .diff import Differ, Delta  # for field constants, formatting


class Report:
    def __init__(self, header):
        """Create blank report, with a header.

        :param header: Report header
        :type header: Header
        """
        self._hdr = header
        self._sections = []

    def add_section(self, section):
        self._sections.append(section)

    @property
    def header(self):
        return self._hdr

    def is_empty(self):
        if len(self._sections) == 0:
            return True
        self._total_rows = 0
        for sect in self._sections:
            self._count_rows(sect)
        return self._total_rows == 0

    def _count_rows(self, sect):
        if sect.body is not None:
            self._total_rows += sect.body.nrow
        for subsect in sect._sections:
            self._count_rows(subsect)

    def __iter__(self):
        return iter(self._sections)


class ReportSection(Report):
    """Section within a report, with data.
    """
    def __init__(self, header, body=None):
        """Create new report section, initialized with header and body.

        :param header: The header for the section
        :type header: SectionHeader
        :param body: The body of the section, or None if this is a container for sub-sections
        :type body: Table
        """
        Report.__init__(self, header)
        self._body = body

    @property
    def body(self):
        return self._body


class Header:
    """Base header class.
    """
    def __init__(self, title=''):
        self._kv = []
        self.title = title

    def add(self, key, value):
        self._kv.append((key, value))

    def get(self, key):
        return (v for k, v in self._kv if k == key)

    def __iter__(self):
        return iter(self._kv)

    def as_dict(self):
        return {k: v for k, v in self._kv}


class ReportHeader(Header):
    """Header for entire report.
    """
    pass


class SectionHeader(Header):
    """Header for one section of a report.
    """
    pass


class Table:
    """Table of values.
    """
    def __init__(self, colnames):
        self._colnames = colnames
        self._rows = []
        self._width = len(colnames)
        self._max_col_widths = list(map(len, colnames))

    def add(self, values):
        if len(values) != self._width:
            raise ValueError('expected {:d} values, got {:d}'.format(self._width, len(values)))
        self._rows.append(values)
        for i, v in enumerate(values):
            n = len(str(v))
            if self._max_col_widths[i] < n:
                self._max_col_widths[i] = n

    def sortby(self, name_or_index):
        name, index = None, None
        if isinstance(name_or_index, int):
            index = name_or_index
        else:
            name = name_or_index
        if name is not None:
            try:
                colnum = self._colnames.index(name)
            except ValueError:
                raise ValueError('column {} not in {}'.format(name, self._colnames))
        else:
            if index < 0 or index >= self._width:
                raise ValueError('index out of range 0..{:d}'.format(self._width - 1))
            colnum = index
        self._rows.sort(key=itemgetter(colnum))

    def __iter__(self):
        return iter(self._rows)

    @property
    def values(self):
        return [{self._colnames[i]: r[i] for i in range(self._width)}
                for r in self._rows]

    @property
    def column_names(self):
        return self._colnames

    @property
    def column_widths(self):
        return self._max_col_widths

    @property
    def ncol(self):
        return self._width

    @property
    def nrow(self):
        return len(self._rows)


## Exceptions

class ReportBackupError(Exception):
    pass

## Formatting


def css_minify(s):
    #return s.replace('\n', ' ').replace('  ', ' ')
    s = s.replace("{ ", "{")
    s = s.replace(" }", "}")
    return s

# CSS for HTML report output
DEFAULT_CSS = [
    "html { font-family: 'Helvetica Neue', Helvetica, Arial, sans-serif; }",
    "body { margin: 2em;}",
    "table { margin-top: 1em; clear: both; border: 0;}",
    "dl, dt, dd { float: left; }",
    "dl, dt { clear: both; }",
    "dt { width: 8em; font-weight: 700; }",
    "dd { width: 32em; }",
    "tr:nth-child(even) { background-color: #E9E9E9; }",
    "tr:nth-child(odd) { background-color: #E9E9E9; }",
    "th, td {padding: 0.2em 0.5em;}",
    "th { text-align: left;  color: black; margin: 0; font-weight: 300;}",
    "h1, h2, h3 { clear: both; margin: 0; padding: 0; }",
    "h1 { font-size: 18; color: rgb(44, 62, 80); }",
    "h2 { font-size: 14; color: black; }"
]


class HTMLFormatter:
    """Format a report as HTML.
    """
    def __init__(self, line_sep='\n', id_column=0, css=None):
        self._sep = line_sep
        self._idcol = id_column
        if css is None:
            css = DEFAULT_CSS
        self._css = css

    def format(self, report):
        text = []
        text.append('<!DOCTYPE html>')
        text.append('<html>')
        text.append('<title>{}</title>'.format(report.header.title))
        text.append('<head>')
        if self._css:
            text.append('<style>')
            text.append("\r\n".join(self._css))
            text.append('</style>')
        text.append('</head>')
        text.append('<body>')
        text.append('<h1>{}</h1>'.format(report.header.title))
        text.append('<dl class="rptmeta">')
        for key, value in report.header:
            text.append('<dt>{}</dt>'.format(key))
            text.append('<dd>{}</dd>'.format(value))
        text.append('</dl>')
        for section in report:
            text.append('<h2>{}</h2>'.format(section.header.title))
            text.append('<dl class="sectmeta">')
            for key, value in section.header:
                text.append('<dt>{}</dt>'.format(key))
                text.append('<dd>{}</dd>'.format(value))
            text.append('</dl>')
            for cond_section in section:
                text.append('<h3>{}</h3>'.format(cond_section.header.title))
                text.append('<dl class="subsectmeta">')
                for key, value in cond_section.header:
                    text.append('<dt>{}</dt>'.format(key))
                    text.append('<dd>{}</dd>'.format(value))
                text.append('</dl>')
                text.append('<table>')
                text.append('<tr>')
                for name in cond_section.body.column_names:
                    text.append('<th>{}</th>'.format(name))
                text.append('</tr>')
                prev_key, i = None, 0
                for row in cond_section.body:
                    row = list(row)
                    key = row[self._idcol]
                    if prev_key and key == prev_key:
                        row[self._idcol] = ''
                    else:
                        prev_key = key
                        i += 1
                    rclass = ('even', 'odd')[i % 2]
                    text.append('<tr class="{}">'.format(rclass))
                    for value in row:
                        text.append('<td>{}</td>'.format(value))
                    text.append('</tr>')
                text.append('</table>')
        text.append('</body>')
        text.append('</html>')
        text_str = self._sep.join(text)
        text_str = text_str.replace("**", "<br>")
        return text_str

class JSONFormatter:
    """Format a report as JSON.
    """
    def __init__(self, id_column=0, indent=2):
        self._indent = indent
        self._idcol = id_column

    def format(self, report):
        obj = dict(
            title=report.header.title,
            info=report.header,
            sections=[
                dict(title=s.header.title,
                     info=s.header,
                     conditions=[dict(
                         title=cs.header.title,
                         info=cs.header,
                         violations=cs.body)
                         for cs in s
                     ]
                )
                for s in report
            ]
        )
        return json.dumps(obj, indent=self._indent, cls=MongoJSONEncoder)


class ReportJSONEncoder(MongoJSONEncoder):
    def default(self, o):
        if isinstance(o, Header):
            return o.as_dict()
        elif isinstance(o, Table):
            return o.values
        return MongoJSONEncoder.default(self, o)


class MarkdownFormatter:
    """Format a report as markdown
    """
    def __init__(self, id_column=0):
        self._idcol = id_column

    def _mapdump(self, d):
        return ', '.join((('{}={}'.format(k, v) for k, v in d.items())))

    def _fixed_width(self, values, widths):
        s = ''.join(["{{:{:d}s}}".format(w + 1).format(str(v))
                    for w, v in zip(widths, values)])
        return s

    def format(self, report):
        lines = []
        self._append_heading(lines, 1, report.header.title)
        self._append_info_section(lines, report.header)
        for section in report:
            self._append_heading(lines, 2, section.header.title)
            self._append_info_section(lines, section.header)
            for cond in section:
                self._append_heading(lines, 3, cond.header.title)
                self._append_info_section(lines, cond.header)
                self._append_violations(lines, cond.body)
        return '\n'.join(lines)

    def _append_info_section(self, lines, info):
        if not info:
            return
        infodict = info.as_dict()
        if infodict:
            text = 'Info: {}'.format(self._mapdump(infodict))
            lines.append(text)

    def _append_heading(self, lines, level, title):
        hashes = '#' * level
        text = '\n{} {} {}\n'.format(hashes, title, hashes)
        lines.append(text)

    def _append_violations(self, lines, data):
        lines.append('\nViolations:\n')
        indent = '    '
        lines.append(indent + self._fixed_width(data.column_names, data.column_widths))
        for row in data:
            lines.append(indent + self._fixed_width(row, data.column_widths))


class Emailer(DoesLogging):
    """Send a report to an email recipient.
    """
    def __init__(self, sender='me@localhost', recipients=('you@remote.host',),
                 subject='Report', server='localhost', port=None, **kwargs):
        """Send reports as email.

        :param: sender Sender of the email
        :param: recipients List of _recipients of the email
        :param: subject Email _subject line
        :param: server SMTP server host
        :param: port SMTP server port (None for default)
        """
        DoesLogging.__init__(self, 'mg.emailer')
        self._sender, self._recipients, self._subject = (sender, recipients,
                                                         subject)
        self._server, self._port = server, port
        self._message = ""

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, value):
        self._subject = value

    def send(self, text, fmt):
        """Send the email message.

        :param text: The text to send
        :type text: str
        :param fmt: The name of the format of the text
        :type fmt: str
        :return: Number of recipients it was sent to
        :rtype: int
        """
        main_fmt, sub_fmt = fmt.split('/')
        if sub_fmt.lower() == "text":
            msg = MIMEText(text, "plain")
        elif sub_fmt.lower() == "html":
            msg = MIMEText(text, "html")
        else:
            raise ValueError("Unknown message format: {}".format(fmt))
        msg['Subject'] = self._subject
        msg['From'] = self._sender
        msg['To'] = ', '.join(self._recipients)
        if self._port is None:
            conn_kwargs = dict(host=self._server)
        else:
            conn_kwargs = dict(host=self._server, port=self._port)
        self._log.info("connect to email server {}".format(conn_kwargs))
        try:
            s = smtplib.SMTP(**conn_kwargs)
            #s.set_debuglevel(2)
            refused = s.sendmail(self._sender, self._recipients, msg.as_string())
            if refused:
                self._log.warn("Email to {:d} recipients was refused".format(len(refused)))
                for person, (code, msg) in refused.items():
                    self._log("Email to {p} was refused ({c}): {m}".format(p=person, c=code, m=msg))
            s.quit()
            n_recip = len(self._recipients)
        except Exception as err:
            self._log.error("connection to SMTP server failed: {}".format(err))
            n_recip = 0
        return n_recip


# ---------------
# Diff formatting
# ---------------

class DiffFormatter:
    """Base class for formatting a 'diff' report.
    """

    TITLE = "Materials Project Database Diff Report"

    def __init__(self, meta, key=None):
        """Constructor.

        :param meta: Report metadata, must have the following keys:
                     - start_time, end_time: string repr of report gen. times
                     - elapsed: float #sec for end_time - start_time
                     - db1, db2: string repr of 2 input database/collections.
        :type meta: dict
        :param key: Record key field
        :type key: str
        """
        self.meta = meta
        self.key = key

    def format(self, result):
        """Format a report from a result object.

        :return: Report body
        :rtype: str
        """
        raise NotImplementedError()

    def result_subsets(self, rs):
        """Break a result set into subsets with the same keys.

        :param rs: Result set, rows of a result as a list of dicts
        :type rs: list of dict
        :return: A set with distinct keys (tuples), and a dict, by these tuples, of max. widths for each column
        """
        keyset, maxwid = set(), {}
        for r in rs:
            key = tuple(sorted(r.keys()))
            keyset.add(key)
            if key not in maxwid:
                maxwid[key] = [len(k) for k in key]
            for i, k in enumerate(key):
                strlen = len("{}".format(r[k]))
                maxwid[key][i] = max(maxwid[key][i], strlen)
        return keyset, maxwid

    def ordered_cols(self, columns, section):
        """Return ordered list of columns, from given columns and the name of the section
        """
        columns = list(columns)  # might be a tuple
        fixed_cols = [self.key]
        if section.lower() == "different":
            fixed_cols.extend([Differ.CHANGED_MATCH_KEY, Differ.CHANGED_OLD, Differ.CHANGED_NEW])
        map(columns.remove, fixed_cols)
        columns.sort()
        return fixed_cols + columns

    def sort_rows(self, rows, section):
        """Sort the rows, as appropriate for the section.

        :param rows: List of tuples (all same length, same values in each position)
        :param section: Name of section, should match const in Differ class
        :return: None; rows are sorted in-place
        """
        #print("@@ SORT ROWS:\n{}".format(rows))
        # Section-specific determination of sort key
        if section.lower() == Differ.CHANGED.lower():
            sort_key = Differ.CHANGED_DELTA
        else:
            sort_key = None
        if sort_key is not None:
            rows.sort(key=itemgetter(sort_key))


class DiffJsonFormatter(DiffFormatter):

    class Encoder(json.JSONEncoder):
        def default(self, o):
            if hasattr(o, 'as_json'):
                return o.as_json()
            return json.JSONEncoder.default(self, o)

#    class Manipulator(pymongo.son_manipulator.SONManipulator):
#        def transform_incoming(self, son, collection):
#            return walk(son, visit_as_json, None)

    def __init__(self, meta, pretty=False, **kwargs):
        """Constructor.

        :param meta: see superclass
        :param pretty: Indented format
        :type pretty: bool
        """
        DiffFormatter.__init__(self, meta, **kwargs)
        self._indent = 4 if pretty else None

    def will_copy(self):
            return True

    def _add_meta(self, result):
        # put metadata into its own section
        result['meta'] = self.meta
        # .. but promote a timestamp, for searching
        result['time'] = self.meta['end_time']

    def format(self, result):
        self._add_meta(result)
        return json.dumps(result, cls=self.Encoder, indent=self._indent)

    def document(self, result):
        """Build dict for MongoDB, expanding result keys as we go.
        """
        self._add_meta(result)
        walker = JsonWalker(JsonWalker.value_json, JsonWalker.dict_expand)
        r = walker.walk(result)
        return r


class DiffHtmlFormatter(DiffFormatter):
    """Format an HTML diff report.
    """
    DIFF_CSS = [
        ".header {padding: 5px; margin: 0 5px;}",
        ".header h1 {color: #165F4B; font-size: 20; text-align: left; margin-left: 20px;}",
        ".header p {color: #666666; margin-left: 20px; height: 12px;}",
        ".header p em {color: #4169E1; font-style: normal;}",
        ".content {padding: 15px; padding-top: 0px; margin: 0; background-color: #F3F3F3;}",
        ".content h2 {color: #2C3E50; font-size: 16px;}",
        ".empty { font-size: 14px; font-style: italic;}",
        ".section {padding: 5px; margin: 10px; background-color: #E2E2E2; border-radius: 5px;}",
        ".section div {margin-left: 10px;}",
        ".section table {margin-left: 5px;}",
        "tr:nth-child(even) { background-color: white; }",
        "tr:nth-child(odd) { background-color: #F5F5F5; }",
        "tr:nth-child(1) { background-color: #778899; font-weight: 500;}",
        "th, td {padding: 0.2em 0.5em;}",
        "th { text-align: left;  color: white; margin: 0;}",
        ".fixed { font-family: Consolas, monaco, monospace; }",
    ]
    css = DEFAULT_CSS + DIFF_CSS

    # for email inlining
    styles = {
        "header": {
            "h1": "color: #165F4B; font-size: 20; text-align: left; margin-left: 20px",
            "p":  "color: #666666; margin-left: 20px; height: 12px",
            "em": "color: #4169E1; font-style: normal"
        },
        "content": {
            "_": "padding: 15px; padding-top: 0px; margin: 0; background-color: #F3F3F3",
            "h2": "color: #2C3E50; font-size: 16px",
            "section": "padding: 5px; margin: 10px; background-color: #E2E2E2; border-radius: 5px"
        },
        "table": {
            "table": "margin-top: 1em; clear: both; border: 0",
            "tr_even": "background-color: white",
            "tr_odd": "background-color: #F5F5F5",
            "tr1": "background-color: #778899; font-weight: 500",
            "th": "text-align: left;  color: white; margin: 0; padding: 0.2em 0.5em",
            "td": "padding: 0.2em 0.5em"
        }
    }

    def __init__(self, meta, url=None, email_mode=False, **kwargs):
        """Constructor.

        :param meta: see superclass
        :param url: Optional URL to create hyperlink for keys
        :type url: str
        """
        DiffFormatter.__init__(self, meta, **kwargs)
        self._url = url
        self._email = email_mode

    def format(self, result):
        """Generate HTML report.

        :return: Report body
        :rtype: str
        """
        css = "\n".join(self.css)
        content = "{}{}".format(self._header(), self._body(result))
        if self._email:
            text = """<!DOCTYPE html>
            <html>
            <div width="100%" style="{sty}">{content}</div>
            </html>
            """.format(css=css, content=content, sty=self.styles["content"]["_"])
        else:
            text = """<html>
            <head><style>{css}</style></head>
            <body>{content}</body>
            </html>
            """.format(css=css, content=content)
        return text

    def _header(self):
        lines = ["<div class='header'><h1{{sh1}}>{t}</h1>".format(t=self.TITLE),
                 "<p{sp}>Compared <em{sem}>{{db1}}</em> with <em{sem}>{{db2}}</em></p>",
                 "<p{sp}>Filter: <span class='fixed'>{{filter}}</span></p>",
                 "<p{sp}>Run time: <em{sem}>{{start_time}}</em> to <em{sem}>{{end_time}}</em> ",
                 "(<em{sem}>{{elapsed:.1f}}</em> sec)</p>",
                 "</div>"]
        if self._email:  # inline the styles
            _c = "header"
            _f = lambda s: s.format(sp=self.style(_c, "p"), sem=self.style(_c, "em"), sh1=self.style(_c, "h1"))
        else:
            _f = lambda s: s.format(sp="", sem="", sh1="")
        lines = map(_f, lines)
        s = "\n".join(lines)
        return s.format(**self.meta)

    def style(self, css_class, elt):
        s = ""
        if css_class in self.styles and elt in self.styles[css_class]:
            s = " style='{}'".format(self.styles[css_class][elt])
        return s

    def _body(self, result):
        body = ["<div class='content'>"]
        for section in result.keys():
            body.append("<div class='section'{{ssec}}><h2{{sh2}}>{t}</h2>".format(t=section.title()))
            if len(result[section]) == 0:
                body.append("<div class='empty'>Empty</div>")
            else:
                body.extend(self._table(section, result[section]))
            body.append("</div>")
        body.append("</div>")
        if self._email:
            _c = "content"
            _f = lambda s: s.format(s_=self.style(_c, "_"), sh2=self.style(_c, "h2"), ssec=self.style(_c, "section"))
        else:
            _f = lambda s: s.format(s_="", sh2="", ssec="")
        body = map(_f, body)
        return '\n'.join(body)

    def _table(self, section, rows):
        if self._email:
            inline = {k: self.style("table", k) for k in self.styles["table"]}
        else:
            inline = dict.fromkeys(self.styles["table"], "")
        subsets, _ = self.result_subsets(rows)
        tables = []
        for subset in subsets:
            tables.append("<table{table}>".format(**inline))
            cols = self.ordered_cols(subset, section)
            # Format the table.
            tables.extend(["<tr{tr1}>".format(**inline)] +
                          ["<th{th}>{c}</th>".format(c=c, **inline) for c in cols] +
                          ["</tr>"])
            self.sort_rows(rows, section)
            for i, r in enumerate(rows):
                tr = "{{tr_{}}}".format(("even", "odd")[i % 2])
                if tuple(sorted(r.keys())) != subset:
                    continue
                if self._url is not None:
                    r[cols[0]] = "<a href='{p}{v}'>{v}</a>".format(p=self._url, v=r[cols[0]])
                tables.extend(["<tr{}>".format(tr).format(**inline)] +
                              ["<td{td}>{d}</td>".format(d=r[c], **inline) for c in cols] +
                              ["</tr>"])
            tables.append("</table>")
        return tables


class DiffTextFormatter(DiffFormatter):
    """Format a plain-text diff report.
    """

    def format(self, result):
        """Generate plain text report.

        :return: Report body
        :rtype: str
        """
        m = self.meta
        lines = ['-' * len(self.TITLE),
                 self.TITLE,
                 '-' * len(self.TITLE),
                 "Compared: {db1} <-> {db2}".format(**m),
                 "Filter: {filter}".format(**m),
                 "Run time: {start_time} -- {end_time} ({elapsed:.1f} sec)".format(**m),
                 ""]
        for section in result.keys():
            lines.append("* " + section.title())
            indent = " " * 4
            if len(result[section]) == 0:
                lines.append("{}EMPTY".format(indent))
            else:
                keyset, maxwid = self.result_subsets(result[section])
                for columns in keyset:
                    ocol = self.ordered_cols(columns, section)
                    mw = maxwid[columns]
                    mw_i = [columns.index(c) for c in ocol]  # reorder indexes
                    fmt = '  '.join(["{{:{:d}s}}".format(mw[i]) for i in mw_i])
                    lines.append("")
                    lines.append(indent + fmt.format(*ocol))
                    lines.append(indent + '-_' * (sum(mw)/2 + len(columns)))
                    rows = result[section]
                    self.sort_rows(rows, section)
                    for r in rows:
                        key = tuple(sorted(r.keys()))
                        if key == columns:
                            values = [str(r[k]) for k in ocol]
                            lines.append(indent + fmt.format(*values))
        return '\n'.join(lines)

    def _record(self, rec):
        fields = ['{}: {}'.format(k, v) for k, v in rec.items()]
        return '{' + ', '.join(fields) + '}'



#!/usr/bin/env python

"""
Created on Jun 19, 2012
"""

from __future__ import division

import unittest
import os
import warnings

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from pymatgen.apps.borg.queen import BorgQueen
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.core.structure import Structure

from matgendb.query_engine import QueryEngine
from matgendb.creator import VaspToDbTaskDrone
from matgendb.tests import common

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Jun 19, 2012"

test_dir = os.path.join(os.path.dirname(__file__), "..", "..",
                        'test_files')

has_mongo = common.has_mongo()


class VaspToDbTaskDroneTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if has_mongo:
            try:
                cls.conn = MongoClient()
            except ConnectionFailure:
                cls.conn = None
        else:
            cls.conn = None

    @unittest.skipUnless(has_mongo, "MongoDB connection required")
    def test_get_valid_paths(self):
        drone = VaspToDbTaskDrone(simulate_mode=True)
        all_paths = []
        for path in os.walk(os.path.join(test_dir, 'db_test')):
            all_paths.extend(drone.get_valid_paths(path))
        self.assertEqual(len(all_paths), 6)

    @unittest.skipUnless(has_mongo, "MongoDB connection required")
    def test_to_from_dict(self):
        drone = VaspToDbTaskDrone(database="wacky", simulate_mode=True)
        d = drone.as_dict()
        drone = VaspToDbTaskDrone.from_dict(d)
        self.assertTrue(drone.simulate)
        self.assertEqual(drone.database, "wacky")

    @unittest.skipUnless(has_mongo, "MongoDB connection required")
    def test_assimilate(self):
        """Borg assimilation code.
        This takes too long for a unit test!
        """
        simulate = True if VaspToDbTaskDroneTest.conn is None else False
        drone = VaspToDbTaskDrone(database="creator_unittest",
                                  simulate_mode=simulate,
                                  parse_dos=True, compress_dos=1)
        queen = BorgQueen(drone)
        queen.serial_assimilate(os.path.join(test_dir, 'db_test'))
        data = queen.get_data()
        self.assertEqual(len(data), 6)
        if VaspToDbTaskDroneTest.conn:
            db = VaspToDbTaskDroneTest.conn["creator_unittest"]
            data = db.tasks.find()
            self.assertEqual(data.count(), 6)
            warnings.warn("Actual db insertion mode.")

        for d in data:
            dir_name = d['dir_name']
            if dir_name.endswith("killed_mp_aflow"):
                self.assertEqual(d['state'], "killed")
                self.assertFalse(d['is_hubbard'])
                self.assertEqual(d['pretty_formula'], "SiO2")
            elif dir_name.endswith("stopped_mp_aflow"):
                self.assertEqual(d['state'], "stopped")
                self.assertEqual(d['pretty_formula'], "ThFe5P3")
            elif dir_name.endswith("success_mp_aflow"):
                self.assertEqual(d['state'], "successful")
                self.assertEqual(d['pretty_formula'], "TbZn(BO2)5")
                self.assertAlmostEqual(d['output']['final_energy'],
                                       -526.66747274, 4)
            elif dir_name.endswith("Li2O_aflow"):
                self.assertEqual(d['state'], "successful")
                self.assertEqual(d['pretty_formula'], "Li2O")
                self.assertAlmostEqual(d['output']['final_energy'],
                                       -14.31446494, 6)
                self.assertEqual(len(d["calculations"]), 2)
                self.assertEqual(d['input']['is_lasph'], False)
                self.assertEqual(d['input']['xc_override'], None)
                self.assertEqual(d["oxide_type"], "oxide")
            elif dir_name.endswith("Li2O"):
                self.assertEqual(d['state'], "successful")
                self.assertEqual(d['pretty_formula'], "Li2O")
                self.assertAlmostEqual(d['output']['final_energy'],
                                       -14.31337758, 6)
                self.assertEqual(len(d["calculations"]), 1)
                self.assertEqual(len(d["custodian"]), 1)
                self.assertEqual(len(d["custodian"][0]["corrections"]), 1)
            elif dir_name.endswith("Li2O_aflow_lasph"):
                self.assertEqual(d['state'], "successful")
                self.assertEqual(d['pretty_formula'], "Li2O")
                self.assertAlmostEqual(d['output']['final_energy'],
                                       -13.998171, 6)
                self.assertEqual(len(d["calculations"]), 2)
                self.assertEqual(d['input']['is_lasph'], True)
                self.assertEqual(d['input']['xc_override'], "PS")

        if VaspToDbTaskDroneTest.conn:
            warnings.warn("Testing query engine mode.")
            qe = QueryEngine(database="creator_unittest")
            self.assertEqual(qe.query().count(), 6)
            #Test mappings by query engine.
            for r in qe.query(criteria={"pretty_formula": "Li2O"},
                              properties=["dir_name", "energy",
                                          "calculations", "input"]):
                if r["dir_name"].endswith("Li2O_aflow"):
                    self.assertAlmostEqual(r['energy'], -14.31446494, 4)
                    self.assertEqual(len(r["calculations"]), 2)
                    self.assertEqual(r["input"]["is_lasph"], False)
                    self.assertEqual(r['input']['xc_override'], None)
                    self.assertEqual(d["oxide_type"], "oxide")
                elif r["dir_name"].endswith("Li2O"):
                    self.assertAlmostEqual(r['energy'],
                                           -14.31337758, 4)
                    self.assertEqual(len(r["calculations"]), 1)
                    self.assertEqual(r["input"]["is_lasph"], False)
                    self.assertEqual(r['input']['xc_override'], None)

            # Test lasph
            e = qe.get_entries({"dir_name":{"$regex":"lasph"}})
            self.assertEqual(len(e), 1)
            self.assertEqual(e[0].parameters["is_lasph"], True)
            self.assertEqual(e[0].parameters["xc_override"], "PS")

            # Test query one.
            d = qe.query_one(criteria={"pretty_formula": "TbZn(BO2)5"},
                             properties=["energy"])
            self.assertAlmostEqual(d['energy'], -526.66747274, 4)

            d = qe.get_entries_in_system(["Li", "O"])
            self.assertEqual(len(d), 3)
            self.assertIsInstance(d[0], ComputedEntry)
            self.assertEqual(d[0].data["oxide_type"], "oxide")

            s = qe.get_structure_from_id(d[0].entry_id)
            self.assertIsInstance(s, Structure)
            self.assertEqual(s.formula, "Li2 O1")

            self.assertIsInstance(qe.get_dos_from_id(d[0].entry_id), CompleteDos)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.drop_database("creator_unittest")


if __name__ == "__main__":
    unittest.main()
